# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Utils for testing."""

from __future__ import annotations

from functools import reduce
from typing import Any, Dict, List, Optional

import yaml
from typing_extensions import TypeAlias

NestedDict: TypeAlias = Dict[str, Any]


def merge_dict(
    a: NestedDict, b: NestedDict, path: Optional[List[str]] = None
) -> NestedDict:
    """Merge two dictionaries into one."""
    if path is None:
        path = []

    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dict(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                raise RuntimeError(f"Conflict at {'.'.join(path + [str(key)])}")
        else:
            a[key] = b[key]
    return a


def merge_dicts(dicts: List[NestedDict]) -> NestedDict:
    """Merge multiple dicts to one."""
    d: NestedDict = {}
    d = reduce(merge_dict, [d, *dicts])
    return d


def merge_yaml_strings(strings: List[str]) -> str:
    """Merge YAML strings into one.

    Args:
        strings (List[str]): YAML strings to merge into one.

    Returns:
        str: A merged YAML string

    """
    dicts = [yaml.safe_load(s) or {} for s in strings]
    d = merge_dicts(dicts)
    return yaml.safe_dump(d)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Context management."""

from __future__ import annotations

from pathlib import Path
from typing import Optional

import friendli
from friendli.utils.fs import get_friendli_directory

team_context_path = get_friendli_directory() / "team"
project_context_path = get_friendli_directory() / "project"


def get_current_team_id() -> Optional[str]:
    """Get team ID of the current context."""
    if friendli.team_id:
        return friendli.team_id

    if not team_context_path.exists():
        return None

    with open(team_context_path, "r", encoding="utf-8") as f:
        group_id = f.read()
        return group_id


def set_current_team_id(team_id: str):
    """Set the current team context."""
    with open(team_context_path, "w", encoding="utf-8") as f:
        f.write(str(team_id))


def get_current_project_id() -> Optional[str]:
    """Get project ID of the current context."""
    if friendli.project_id:
        return friendli.project_id

    if not project_context_path.exists():
        return None

    with open(project_context_path, "r", encoding="utf-8") as f:
        project_id = f.read()
        return project_id


def set_current_project_id(pf_project_id: str):
    """Set the current project context."""
    with open(project_context_path, "w", encoding="utf-8") as f:
        f.write(str(pf_project_id))


def clear_contexts():
    """Clear all contexts."""
    Path.unlink(team_context_path, missing_ok=True)
    Path.unlink(project_context_path, missing_ok=True)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Endpoint GQL Clients."""

from __future__ import annotations

from typing import Any, Dict, List

from friendli.client.graphql.base import GqlClient

CreateEndpointOp = """
mutation CreateEndpointWithHfModel($input: DedicatedEndpointCreateWithHfInput!) {
    dedicatedEndpointCreateWithHf(input: $input) {
        ... on UserPermissionError {
            message
        }
        ... on TeamNotExistError {
            message
        }
        ... on DedicatedEndpointCreateWithHfSuccess {
            endpoint {
                id
                name
                numGpu
                gpuType
                hfModelRepo
                status
                endpointUrl
                createdBy {
                    email
                    id
                    name
                }
                createdAt
                updatedAt
            }
        }
    }
}
"""

ListEndpointsOp = """
query ListDedicatedEndpointsInProject($id: ID!) {
    dedicatedProject(id: $id) {
        id
        name
        endpoints {
            totalCount
            edges {
                node {
                    id
                    name
                    hfModelRepo
                    gpuType
                    numGpu
                    hfModelRepo
                    status
                    phase {
                        ... on DedicatedEndpointPhaseRunning {
                            msg
                            desiredReplica
                            currReplica
                        }
                        ... on DedicatedEndpointPhaseInitializing {
                            step
                            msg
                        }
                    }
                    endpointUrl
                    createdBy {
                        email
                        id
                        name
                    }
                    advancedConfig {
                        maxBatchSize
                        autoscalingMin
                        autoscalingMax
                    }
                    createdAt
                    updatedAt
                }
            }
        }
    }
}
"""

GetEndpointOp = """
query GetDedicatedEndpoint($dedicatedEndpointId: ID!) {
    dedicatedEndpoint(id: $dedicatedEndpointId) {
        id
        name
        hfModelRepo
        gpuType
        numGpu
        status
        phase {
            ... on DedicatedEndpointPhaseInitializing {
                msg
                step
            }
            ... on DedicatedEndpointPhaseRunning {
                currReplica
                msg
                desiredReplica
            }
        }
        endpointUrl
        createdBy {
            email
            id
            name
        }
        createdAt
        updatedAt
        advancedConfig {
            autoscalingMax
            autoscalingMin
            maxBatchSize
        }
    }
}
"""

TerminateEndpointOp = """
mutation TerminateEndpoint($input: DedicatedEndpointTerminateInput!) {
    dedicatedEndpointTerminate(input: $input) {
        ... on DedicatedEndpointTerminateSuccess {
            endpoint {
                id
            }
        }
        ... on UserPermissionError {
            message
        }
        ... on TeamNotExistError {
            message
        }
    }
}
"""

ListInstancesOp = """
query DedicatedInstanceList {
    dedicatedInstanceList {
        id
        name
        disabledReason
        disabled
        options {
            quantity
            pricePerHour
            id
        }
    }
}
"""


class EndpointGqlClient(GqlClient):
    """Endpoint GQL client."""

    def create(  # pylint: disable=too-many-arguments
        self,
        team_id: str,
        project_id: str,
        name: str,
        model_repo: str,
        gpu_type: str,
        num_gpus: int,
    ) -> Dict[str, Any]:
        """Get current user info."""
        instance_option_id = self._get_instance_option_id(
            gpu_type=gpu_type, num_gpus=num_gpus
        )
        response = self.run(
            query=CreateEndpointOp,
            variables={
                "input": {
                    "teamId": team_id,
                    "projectId": project_id,
                    "name": name,
                    "modelRepo": model_repo,
                    "instanceOptionId": instance_option_id,
                }
            },
        )
        return response["dedicatedEndpointCreateWithHf"]["endpoint"]

    def list(self, project_id: str) -> List[Dict[str, Any]]:
        """List team projects."""
        response = self.run(query=ListEndpointsOp, variables={"id": project_id})
        return response["dedicatedProject"]["endpoints"]["edges"]

    def get(self, endpoint_id: str) -> Dict[str, Any]:
        """Get endpoint data."""
        response = self.run(
            query=GetEndpointOp,
            variables={"dedicatedEndpointId": endpoint_id},
        )
        return response["dedicatedEndpoint"]

    def terminate(self, team_id: str, project_id: str, endpoint_id: str) -> None:
        """Terminate a running endpoint."""
        self.run(
            query=TerminateEndpointOp,
            variables={
                "input": {
                    "teamId": team_id,
                    "projectId": project_id,
                    "endpointId": endpoint_id,
                }
            },
        )

    def _get_instance_option_id(self, gpu_type: str, num_gpus: int) -> str:
        response = self.run(query=ListInstancesOp)
        instances: list = response["dedicatedInstanceList"]

        available_gpus = []
        for instance in instances:
            name = instance["name"]
            if name == gpu_type:
                available_quantities = []
                for option in instance["options"]:
                    quantity = option["quantity"]
                    if quantity == num_gpus:
                        return option["id"]
                    available_quantities.append(quantity)

                raise ValueError(
                    f"'{gpu_type}' cannot be used with the quantity of '{num_gpus}'. "
                    f"Supported number of GPUs are {', '.join(available_quantities)}"
                )
            available_gpus.append(name)

        raise ValueError(
            f"GPU '{gpu_type}' is not supported. "
            f"Supported GPU types are {', '.join(name)}"
        )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli GQL Client Service."""

from __future__ import annotations

import time
from typing import Any, Dict, Optional

import httpx
from gql import Client, gql
from gql.transport.exceptions import TransportError, TransportQueryError
from gql.transport.httpx import HTTPXTransport

from friendli.auth import TokenType, get_auth_header, get_token, update_token
from friendli.di.injector import get_injector
from friendli.errors import APIError, AuthenticationError, AuthorizationError
from friendli.utils.url import URLProvider


def get_default_gql_client(url: Optional[str] = None) -> Client:
    """Get a default GraphQL client."""
    if url is None:
        injector = get_injector()
        url_provider = injector.get(URLProvider)
        url = url_provider.get_web_backend_uri("api/graphql")

    gql_transport = HTTPXTransport(url=url, headers=get_auth_header())
    gql_client = Client(transport=gql_transport, fetch_schema_from_transport=True)
    return gql_client


class GqlClient:
    """Base interface of graphql client to Friendli system."""

    def __init__(
        self,
        url: Optional[str] = None,
        max_retries: int = 3,
        retry_interval: float = 1,
    ) -> None:
        """Initialize GqlClient."""
        self._client = get_default_gql_client(url)
        self._max_retries = max_retries
        self._retry_interval = retry_interval

    def run(
        self, query: str, variables: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """Run graphql."""
        attempts = 0
        while attempts < self._max_retries:
            try:
                result = self._client.execute(
                    document=gql(query), variable_values=variables
                )
                return result
            except TransportQueryError as exc:
                errors = exc.errors
                if errors:
                    msg = errors[0]["message"]
                    if msg == "Operation only allowed for authenticated users.":
                        attempts += 1
                        if attempts == self._max_retries:
                            raise APIError(detail=msg) from exc
                        self._token_refresh()
                        time.sleep(self._retry_interval)
                        continue
                    raise APIError(detail=msg) from exc
                raise APIError(detail=str(exc)) from exc
            except TransportError as exc:
                attempts += 1
                if attempts == self._max_retries:
                    raise APIError(detail=str(exc)) from exc
                time.sleep(self._retry_interval)

        raise APIError("Request failed.")

    def _token_refresh(self) -> None:
        refresh_token = get_token(TokenType.REFRESH)
        if refresh_token is None:
            raise AuthorizationError(
                "Failed to refresh access token. Please login again."
            )

        injector = get_injector()
        url_provider = injector.get(URLProvider)
        refresh_url = url_provider.get_web_backend_uri("/api/auth/session/refresh")

        response = httpx.post(
            refresh_url,
            headers={
                **get_auth_header(token=refresh_token),
                "Content-Type": "application/json",
                "Accept": "application/json",
                "rid": "anti-csrf",
                "st-auth-mode": "header",
            },
        )
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            raise AuthenticationError(
                "Failed to refresh token. Please login again."
            ) from exc
        headers = response.headers
        access_token = headers["st-access-token"]
        refresh_token = headers["st-refresh-token"]
        update_token(TokenType.ACCESS, access_token)
        update_token(TokenType.REFRESH, refresh_token)
        # Initialize GQL client again with the updated auth headers.
        self._client = get_default_gql_client()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Serving API Interface."""

# pylint: disable=no-name-in-module

from __future__ import annotations

import json
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, Optional, Type, TypeVar, Union

import grpc
import grpc._channel
import grpc.aio
import grpc.aio._call
import grpc.aio._channel
import httpx
from google.protobuf import json_format
from google.protobuf import message as pb_message
from pydantic import BaseModel
from typing_extensions import Self

from friendli.auth import get_auth_header
from friendli.errors import APIError
from friendli.utils.request import (
    DEFAULT_CONNECTION_LIMITS,
    DEFAULT_REQ_TIMEOUT,
    DEFAULT_TIMEOUT,
    transform_request_data,
)

_GenerationLine = TypeVar("_GenerationLine", bound=BaseModel)


class GenerationStream(ABC, Generic[_GenerationLine]):
    """Generation stream."""

    def __init__(self, response: httpx.Response) -> None:
        """Initializes generation stream."""
        self._response = response
        self._iter = response.iter_lines()

    def __iter__(self) -> Self:  # noqa: D105
        return self

    @abstractmethod
    def __next__(self) -> _GenerationLine:
        """Iterates the stream."""


class AsyncGenerationStream(ABC, Generic[_GenerationLine]):
    """Asynchronous generation stream."""

    def __init__(self, response: httpx.Response) -> None:
        """Initializes generation stream."""
        self._response = response
        self._iter = response.aiter_lines()

    def __aiter__(self) -> Self:  # noqa: D105
        return self

    @abstractmethod
    async def __anext__(self) -> _GenerationLine:
        """Iterates the stream."""


class GrpcGenerationStream(ABC, Generic[_GenerationLine]):
    """Generation stream."""

    def __init__(self, response: grpc._channel._MultiThreadedRendezvous) -> None:
        """Initializes generation stream."""
        self._iter = response

    def __iter__(self) -> Self:  # noqa: D105
        return self

    @abstractmethod
    def __next__(self) -> _GenerationLine:
        """Iterates the stream."""


class AsyncGrpcGenerationStream(ABC, Generic[_GenerationLine]):
    """Generation stream."""

    def __init__(self, response: grpc.aio._call.UnaryStreamCall) -> None:
        """Initializes generation stream."""
        self._iter = response.__aiter__()  # type: ignore

    def __aiter__(self) -> Self:  # noqa: D105
        return self

    @abstractmethod
    async def __anext__(self) -> _GenerationLine:
        """Iterates the stream."""


_HttpxClient = TypeVar("_HttpxClient", bound=Union[httpx.Client, httpx.AsyncClient])
_ProtoMsgType = TypeVar("_ProtoMsgType", bound=Type[pb_message.Message])


class _DefaultHttpxClient(httpx.Client):
    def __init__(self, **kwargs: Any) -> None:
        kwargs.setdefault("timeout", DEFAULT_TIMEOUT)
        kwargs.setdefault("limits", DEFAULT_CONNECTION_LIMITS)
        kwargs.setdefault("follow_redirects", True)
        super().__init__(**kwargs)


class _DefaultAsyncHttpxClient(httpx.AsyncClient):
    def __init__(self, **kwargs: Any) -> None:
        kwargs.setdefault("timeout", DEFAULT_TIMEOUT)
        kwargs.setdefault("limits", DEFAULT_CONNECTION_LIMITS)
        kwargs.setdefault("follow_redirects", True)
        super().__init__(**kwargs)


class BaseAPI(ABC, Generic[_HttpxClient, _ProtoMsgType]):
    """Base API interface."""

    _http_client: _HttpxClient

    def __init__(
        self,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
    ) -> None:
        """Initializes BaseAPI."""
        self._base_url = base_url
        self._use_protobuf = use_protobuf

    @property
    @abstractmethod
    def _api_path(self) -> str:
        """API URL path."""

    @property
    @abstractmethod
    def _method(self) -> str:
        """API call method."""

    @property
    @abstractmethod
    def _content_type(self) -> str:
        """Request content type."""

    @property
    @abstractmethod
    def _request_pb_cls(self) -> _ProtoMsgType:
        """Protobuf message class to serialize the data of request body."""

    def _build_http_request(self, data: dict[str, Any]) -> httpx.Request:
        """Build request."""
        return self._http_client.build_request(
            method=self._method,
            url=self._build_http_url(),
            content=self._build_content(data),
            files=self._build_files(data),
            headers=self._get_headers(),
        )

    def _build_http_url(self) -> httpx.URL:
        assert self._base_url is not None
        url = os.path.join(self._base_url, self._api_path)
        return httpx.URL(url)

    def _build_grpc_url(self) -> str:
        if self._base_url is None:
            raise ValueError(
                "You need to provide the gRPC server address through `base_url`."
            )
        return self._base_url

    def _get_headers(self) -> Dict[str, Any]:
        return {
            "Content-Type": self._content_type,
            **get_auth_header(),
        }

    def _build_files(self, data: dict[str, Any]) -> dict[str, Any] | None:
        if self._content_type.startswith("multipart/form-data"):
            files = {}
            for key, val in data.items():
                if val is not None:
                    files[key] = (None, val)
            return files
        return None

    def _build_content(self, data: dict[str, Any]) -> bytes | None:
        if self._content_type.startswith("multipart/form-data"):
            return None

        if self._use_protobuf:
            pb_cls = self._request_pb_cls
            request_pb = pb_cls()
            json_format.ParseDict(data, request_pb)
            return request_pb.SerializeToString()

        return json.dumps(data).encode()

    def _build_grpc_request(self, data: dict[str, Any]) -> pb_message.Message:
        pb_cls = self._request_pb_cls
        return pb_cls(**data)


class ServingAPI(BaseAPI[httpx.Client, _ProtoMsgType]):
    """Serving API interface."""

    def __init__(
        self,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
        use_grpc: bool = False,
        http_client: Optional[httpx.Client] = None,
        grpc_channel: Optional[grpc.Channel] = None,
    ) -> None:
        """Initializes ServingAPI."""
        super().__init__(
            base_url=base_url,
            use_protobuf=use_protobuf,
        )

        self._use_grpc = use_grpc
        self._http_client = http_client or _DefaultHttpxClient()
        self._grpc_channel = grpc_channel
        self._grpc_stub = None

    def __enter__(self) -> ServingAPI:
        """Enter the context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the context manager with cleaning up resources."""
        self.close()

    def close(self) -> None:
        """Close the gRPC channel and HTTP client."""
        if self._grpc_channel:
            self._grpc_channel.close()
        self._http_client.close()

    def _get_grpc_stub(self, channel: grpc.Channel) -> Any:
        raise NotImplementedError  # pragma: no cover

    def _request(self, *, data: dict[str, Any], stream: bool) -> Any:
        # TODO: Add retry / handle timeout and etc.
        data = transform_request_data(data)

        if self._use_grpc:
            grpc_request = self._build_grpc_request(data=data)
            if not self._grpc_channel:
                self._grpc_channel = grpc.insecure_channel(self._build_grpc_url())
            try:
                if not self._grpc_stub:
                    self._grpc_stub = self._get_grpc_stub(self._grpc_channel)
            except NotImplementedError as exc:
                raise ValueError("This API does not support gRPC.") from exc
            assert self._grpc_stub
            grpc_response = self._grpc_stub.Generate(grpc_request)
            return grpc_response

        http_request = self._build_http_request(data=data)
        http_response = self._http_client.send(request=http_request, stream=stream)
        self._check_http_error(http_response)
        return http_response

    def _check_http_error(self, response: httpx.Response) -> None:
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            if response.status_code == 404:
                endpoint_url = self._build_http_url()
                raise APIError(
                    f"Endpoint ({endpoint_url}) is not found. This may be due to an "
                    "invalid model name or endpoint ID. For serverless endpoints, see "
                    "https://docs.friendli.ai/guides/serverless_endpoints/pricing "
                    "to find out availble models. For dedicated endpoints, check your "
                    "endpoiont ID again."
                ) from exc

            resp_content = response.read()
            raise APIError(resp_content.decode()) from exc


class AsyncServingAPI(BaseAPI[httpx.AsyncClient, _ProtoMsgType]):
    """Async Serving API interface."""

    def __init__(
        self,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
        use_grpc: bool = False,
        http_client: Optional[httpx.AsyncClient] = None,
        grpc_channel: Optional[grpc.aio.Channel] = None,
    ) -> None:
        """Initializes AsyncServingAPI."""
        super().__init__(base_url=base_url, use_protobuf=use_protobuf)

        self._use_grpc = use_grpc
        self._http_client = http_client or _DefaultAsyncHttpxClient()
        self._grpc_channel = grpc_channel
        self._grpc_stub = None

    async def __aenter__(self) -> AsyncServingAPI:
        """Enter the async context manager."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the async context manager with cleaning up resources."""
        await self.close()

    async def close(self) -> None:
        """Close the gRPC channel and HTTP client."""
        if self._grpc_channel:
            await self._grpc_channel.close(grace=None)
        await self._http_client.aclose()

    def _get_grpc_stub(self, channel: grpc.aio.Channel) -> Any:
        raise NotImplementedError  # pragma: no cover

    async def _request(self, *, data: dict[str, Any], stream: bool) -> Any:
        # TODO: Add retry / handle timeout and etc.
        data = transform_request_data(data)

        if self._use_grpc:
            grpc_request = self._build_grpc_request(data=data)
            if not self._grpc_channel:
                self._grpc_channel = grpc.aio.insecure_channel(self._build_grpc_url())
            try:
                if not self._grpc_stub:
                    self._grpc_stub = self._get_grpc_stub(self._grpc_channel)
            except NotImplementedError as exc:
                raise ValueError("This API does not support gRPC.") from exc
            assert self._grpc_stub
            grpc_response = self._grpc_stub.Generate(
                grpc_request, timeout=DEFAULT_REQ_TIMEOUT
            )
            return grpc_response

        http_request = self._build_http_request(data=data)
        http_response = await self._http_client.send(
            request=http_request, stream=stream
        )
        await self._check_http_error(http_response)

        return http_response

    async def _check_http_error(self, response: httpx.Response) -> None:
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            if response.status_code == 404:
                raise APIError(
                    "Endpoint is not found. This may be due to an invalid model name. "
                    "See https://docs.friendli.ai/guides/serverless_endpoints/pricing "
                    "to find out availble models."
                ) from exc

            resp_content = await response.aread()
            raise APIError(resp_content.decode()) from exc



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli client enums."""


from __future__ import annotations

from enum import Enum


class ModelDataType(str, Enum):
    """Model dtype enums."""

    BF16 = "bf16"
    FP16 = "fp16"
    FP32 = "fp32"
    FP8_E4M3 = "fp8_e4m3"
    INT8 = "int8"
    INT4 = "int4"


class CheckpointFileType(str, Enum):
    """Checkpoint file types."""

    HDF5 = "hdf5"
    SAFETENSORS = "safetensors"

    def __str__(self):
        """Convert to a human-readable string."""
        return self.value


class FileSizeType(str, Enum):
    """File size type."""

    LARGE = "LARGE"
    SMALL = "SMALL"


class QuantMode(str, Enum):
    """Supported quantization modes."""

    SMOOTH_QUANT = "smoothquant"
    AWQ = "awq"
    FP8 = "fp8"


class QuantDatasetFormat(str, Enum):
    """Supported file format for calibration datasets for quantization."""

    JSON = "json"
    CSV = "csv"
    PARQUET = "parquet"
    TXT = "txt"


class ResponseFormat(str, Enum):
    """Response formats of text-to-image."""

    URL = "url"
    PNG = "png"
    JPEG = "jpeg"
    RAW = "raw"



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Friendli Images API."""

from __future__ import annotations

from typing import Optional

import httpx

from friendli.sdk.api.images.text_to_image import AsyncTextToImage, TextToImage


class Images:
    """Images API."""

    text_to_image: TextToImage

    def __init__(
        self,
        base_url: Optional[str] = None,
        http_client: Optional[httpx.Client] = None,
    ) -> None:
        """Initialize Images."""
        self.text_to_image = TextToImage(base_url=base_url, http_client=http_client)

    def close(self) -> None:
        """Clean up all clients' resources."""
        self.text_to_image.close()


class AsyncImages:
    """Asynchronous images API."""

    text_to_image: AsyncTextToImage

    def __init__(
        self,
        base_url: Optional[str] = None,
        http_client: Optional[httpx.AsyncClient] = None,
    ) -> None:
        """Initialize Images."""
        self.text_to_image = AsyncTextToImage(
            base_url=base_url, http_client=http_client
        )

    async def close(self) -> None:
        """Clean up all clients' resources."""
        await self.text_to_image.close()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI API Request Utilities."""

from __future__ import annotations

from typing import Any

import httpx
import pydantic
from requests.exceptions import HTTPError

from friendli.utils.compat import model_dump
from friendli.utils.url import discuss_url

DEFAULT_REQ_TIMEOUT = 600.0
MAX_RETRIES = 3
DEFAULT_TIMEOUT = httpx.Timeout(timeout=DEFAULT_REQ_TIMEOUT, connect=5.0)
DEFAULT_CONNECTION_LIMITS = httpx.Limits(
    max_connections=1000, max_keepalive_connections=100
)


def decode_http_err(exc: HTTPError) -> str:
    """Decode HTTP error."""
    try:
        if exc.response.status_code == 500:
            error_str = f"Internal Server Error: Please contact to system admin via {discuss_url}"
        elif exc.response.status_code == 404:
            error_str = (
                "Not Found: The requested resource is not found. Please check it again. "
                f"If you cannot find out why this error occurs, please visit {discuss_url}."
            )
        else:
            response = exc.response
            detail_json = response.json()
            if "detail" in detail_json:
                error_str = f"Error Code: {response.status_code}\nDetail: {detail_json['detail']}"
            elif "error_description" in detail_json:
                error_str = (
                    f"Error Code: {response.status_code}\n"
                    f"Detail: {detail_json['error_description']}"
                )
            else:
                error_str = f"Error Code: {response.status_code}"
    except ValueError:
        error_str = exc.response.content.decode()

    return error_str


def transform_request_data(data: Any) -> Any:
    """Transform the data to be serializable."""
    if isinstance(data, dict):
        return {k: transform_request_data(v) for k, v in data.items()}

    if isinstance(data, list):
        return [transform_request_data(e) for e in data]

    if isinstance(data, pydantic.BaseModel):
        return model_dump(data, exclude_unset=True)

    return data



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Schema of Endpoint Config."""

from __future__ import annotations

from typing import Any, Dict, List, Optional

import pydantic


class GpuConfig(pydantic.BaseModel):
    """GPU config."""

    type: str
    count: int


class EndpointConfig(pydantic.BaseModel):
    """Endpoint config."""

    name: str
    gpu: GpuConfig
    model: str
    adapters: Optional[List[str]] = None
    launch_config: Optional[Dict[str, Any]] = None



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: response_format.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from __future__ import annotations

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x15response_format.proto\x12\x04orca"\x87\x01\n\x0eResponseFormat\x12\'\n\x04type\x18\x01 \x01(\x0e\x32\x19.orca.ResponseFormat.Type\x12\x13\n\x06schema\x18\x02 \x01(\tH\x00\x88\x01\x01",\n\x04Type\x12\x08\n\x04text\x10\x00\x12\x0f\n\x0bjson_object\x10\x01\x12\t\n\x05regex\x10\x02\x42\t\n\x07_schemab\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "response_format_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_RESPONSEFORMAT"]._serialized_start = 32
    _globals["_RESPONSEFORMAT"]._serialized_end = 167
    _globals["_RESPONSEFORMAT_TYPE"]._serialized_start = 112
    _globals["_RESPONSEFORMAT_TYPE"]._serialized_end = 156
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Commonly used Schemas."""

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel


class Creator(BaseModel):
    """Fine-tuning creator schema."""

    id: Optional[str]
    name: Optional[str]
    email: Optional[str]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli V1 Completion Serving API Schemas."""

from __future__ import annotations

from typing import List, Literal

from pydantic import BaseModel
from typing_extensions import Required, TypeAlias, TypedDict

BeamSearchType: TypeAlias = Literal["DETERMINISTIC", "STOCHASTIC", "NAIVE_SAMPLING"]


class TokenSequenceParam(TypedDict, total=False):
    """Token sequence param schema."""

    tokens: Required[List[int]]


class CompletionChoice(BaseModel):
    """Completion choice schema."""

    index: int
    seed: int
    text: str
    tokens: List[int]


class CompletionUsage(BaseModel):
    """Completion usage schema."""

    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class Completion(BaseModel):
    """Completion schema."""

    choices: List[CompletionChoice]
    usage: CompletionUsage


class CompletionLine(BaseModel):
    """Completion line schema."""

    event: str
    index: int = 0
    text: str = ""
    token: int



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: text_to_image.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from __future__ import annotations

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x13text_to_image.proto"\xd8\x02\n\x14V1TextToImageRequest\x12\x0e\n\x06prompt\x18\x01 \x01(\t\x12\x1c\n\x0fnegative_prompt\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x18\n\x0bnum_outputs\x18\x03 \x01(\x05H\x01\x88\x01\x01\x12 \n\x13num_inference_steps\x18\x04 \x01(\x05H\x02\x88\x01\x01\x12\x1b\n\x0eguidance_scale\x18\x05 \x01(\x02H\x03\x88\x01\x01\x12\x11\n\x04seed\x18\x06 \x01(\x05H\x04\x88\x01\x01\x12\x1c\n\x0fresponse_format\x18\x07 \x01(\tH\x05\x88\x01\x01\x12\x12\n\x05model\x18\x08 \x01(\tH\x06\x88\x01\x01\x42\x12\n\x10_negative_promptB\x0e\n\x0c_num_outputsB\x16\n\x14_num_inference_stepsB\x11\n\x0f_guidance_scaleB\x07\n\x05_seedB\x12\n\x10_response_formatB\x08\n\x06_modelb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "text_to_image_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_V1TEXTTOIMAGEREQUEST"]._serialized_start = 24
    _globals["_V1TEXTTOIMAGEREQUEST"]._serialized_end = 368
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Auth Tools."""

from __future__ import annotations

import functools
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Union

import requests
from requests import HTTPError, JSONDecodeError
from typing_extensions import TypeAlias

import friendli
from friendli.context import get_current_team_id
from friendli.di.injector import get_injector
from friendli.errors import APIError, AuthorizationError
from friendli.logging import logger
from friendli.utils.fs import get_friendli_directory
from friendli.utils.request import DEFAULT_REQ_TIMEOUT, decode_http_err
from friendli.utils.url import URLProvider

access_token_path = get_friendli_directory() / "access_token"
refresh_token_path = get_friendli_directory() / "refresh_token"
mfa_token_path = get_friendli_directory() / "mfa_token"


class TokenType(str, Enum):
    """Token type enums."""

    ACCESS = "ACCESS"
    REFRESH = "REFRESH"
    MFA = "MFA"


token_path_map = {
    TokenType.ACCESS: access_token_path,
    TokenType.REFRESH: refresh_token_path,
    TokenType.MFA: mfa_token_path,
}

ResponseBody: TypeAlias = Union[Dict[str, Any], List[Dict[str, Any]], None]


def get_auth_header(
    token: Optional[str] = None, team_id: Optional[str] = None
) -> Dict[str, Any]:
    """Get authorization header.

    Returns:
        Dict[str, Any]: HTTP Authorization headers for the request.

    """
    token_: Optional[str]

    token_from_cfg = get_token(TokenType.ACCESS)
    token_from_env = friendli.token

    headers = {}
    if token is not None:
        token_ = token
    elif token_from_env:
        if token_from_cfg:
            logger.warning(
                "You've entered your login information in two places - through the "
                "'FRIENDLI_TOKEN' environment variable and the 'friendli login' CLI "
                "command. We will use the access token from the 'FRIENDLI_TOKEN' "
                "environment variable and ignore the login session details. This might "
                "lead to unexpected authorization errors. If you prefer to use the "
                "login session instead, unset the 'FRIENDLI_TOKEN' environment "
                "variable. If you don't want to see this warning again, run "
                "'friendli logout' to remove the login session."
            )
        token_ = token_from_env
    else:
        token_ = token_from_cfg

    if token_ is not None:
        headers["Authorization"] = f"Bearer {token_}"

    team_id = team_id or get_current_team_id()
    if team_id is not None:
        headers["X-Friendli-Team"] = team_id

    return headers


def get_token(token_type: TokenType) -> Union[str, None]:
    """Get an auth token."""
    try:
        if token_type == TokenType.ACCESS:
            return access_token_path.read_text()
        if token_type == TokenType.REFRESH:
            return refresh_token_path.read_text()
        if token_type == TokenType.MFA:
            return mfa_token_path.read_text()
        raise ValueError("Invalid token type.")
    except FileNotFoundError:
        return None


def update_token(token_type: TokenType, token: str) -> None:
    """Update saved token value."""
    token_path_map[token_type].write_text(token)


def delete_token(token_type: TokenType) -> None:
    """Delete token."""
    token_path_map[token_type].unlink(missing_ok=True)


def clear_tokens() -> None:
    """Clear all tokens."""
    for token_type in TokenType:
        delete_token(token_type)


# pylint: disable=too-many-nested-blocks
def safe_request(func: Callable[..., requests.Response]) -> Callable[..., Any]:
    """Decorator for automatic token refresh."""

    @functools.wraps(func)
    def inner(*args, **kwargs) -> Any:
        injector = get_injector()
        url_provider = injector.get(URLProvider)

        resp = func(*args, **kwargs)
        if resp.status_code in (401, 403):
            refresh_token = get_token(TokenType.REFRESH)
            if refresh_token is not None:
                refresh_r = requests.post(
                    url_provider.get_web_backend_uri("/api/auth/cli/refresh_token"),
                    json={"refresh_token": refresh_token},
                    timeout=DEFAULT_REQ_TIMEOUT,
                )
                try:
                    refresh_r.raise_for_status()
                except requests.HTTPError as exc:
                    raise AuthorizationError(
                        "Failed to refresh access token. Please login again."
                    ) from exc

                update_token(
                    token_type=TokenType.ACCESS, token=refresh_r.json()["accessToken"]
                )
                update_token(
                    token_type=TokenType.REFRESH,
                    token=refresh_r.json()["refreshToken"],
                )
                # We need to restore file offset if we want to transfer file objects
                if "files" in kwargs:
                    files = kwargs["files"]
                    for _, file_tuple in files.items():
                        for element in file_tuple:
                            if hasattr(element, "seek"):
                                # Restore file offset
                                element.seek(0)
                resp = func(*args, **kwargs)
                resp.raise_for_status()
            else:
                raise AuthorizationError(
                    "Failed to refresh access token. Please login again."
                )
        else:
            try:
                resp.raise_for_status()
            except HTTPError as exc:
                raise APIError(decode_http_err(exc)) from exc
        try:
            return resp.json()
        except JSONDecodeError:
            return None

    return inner



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

from __future__ import annotations

from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers

from friendli.schema.api.v1.codegen import response_format_pb2 as _response_format_pb2

DESCRIPTOR: _descriptor.FileDescriptor

class ToolCall(_message.Message):
    __slots__ = ("id", "type", "function")

    class Function(_message.Message):
        __slots__ = ("name", "arguments")
        NAME_FIELD_NUMBER: _ClassVar[int]
        ARGUMENTS_FIELD_NUMBER: _ClassVar[int]
        name: str
        arguments: str
        def __init__(
            self, name: _Optional[str] = ..., arguments: _Optional[str] = ...
        ) -> None: ...
    ID_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_FIELD_NUMBER: _ClassVar[int]
    id: str
    type: str
    function: ToolCall.Function
    def __init__(
        self,
        id: _Optional[str] = ...,
        type: _Optional[str] = ...,
        function: _Optional[_Union[ToolCall.Function, _Mapping]] = ...,
    ) -> None: ...

class Message(_message.Message):
    __slots__ = ("content", "role", "name", "tool_call_id", "tool_calls")
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    TOOL_CALL_ID_FIELD_NUMBER: _ClassVar[int]
    TOOL_CALLS_FIELD_NUMBER: _ClassVar[int]
    content: str
    role: str
    name: str
    tool_call_id: str
    tool_calls: _containers.RepeatedCompositeFieldContainer[ToolCall]
    def __init__(
        self,
        content: _Optional[str] = ...,
        role: _Optional[str] = ...,
        name: _Optional[str] = ...,
        tool_call_id: _Optional[str] = ...,
        tool_calls: _Optional[_Iterable[_Union[ToolCall, _Mapping]]] = ...,
    ) -> None: ...

class Tool(_message.Message):
    __slots__ = ("type", "function")

    class Function(_message.Message):
        __slots__ = ("name", "description", "parameters")
        NAME_FIELD_NUMBER: _ClassVar[int]
        DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
        PARAMETERS_FIELD_NUMBER: _ClassVar[int]
        name: str
        description: str
        parameters: _struct_pb2.Struct
        def __init__(
            self,
            name: _Optional[str] = ...,
            description: _Optional[str] = ...,
            parameters: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...,
        ) -> None: ...
    TYPE_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_FIELD_NUMBER: _ClassVar[int]
    type: str
    function: Tool.Function
    def __init__(
        self,
        type: _Optional[str] = ...,
        function: _Optional[_Union[Tool.Function, _Mapping]] = ...,
    ) -> None: ...

class V1ChatCompletionsRequest(_message.Message):
    __slots__ = (
        "messages",
        "model",
        "frequency_penalty",
        "logit_bias",
        "min_tokens",
        "max_tokens",
        "n",
        "presence_penalty",
        "stop",
        "stream",
        "temperature",
        "top_p",
        "timeout_microseconds",
        "logprobs",
        "top_logprobs",
        "top_k",
        "repetition_penalty",
        "seed",
        "eos_token",
        "tools",
        "response_format",
        "tool_choice",
        "parallel_tool_calls",
    )

    class LogitBiasEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: int
        value: float
        def __init__(
            self, key: _Optional[int] = ..., value: _Optional[float] = ...
        ) -> None: ...
    MESSAGES_FIELD_NUMBER: _ClassVar[int]
    MODEL_FIELD_NUMBER: _ClassVar[int]
    FREQUENCY_PENALTY_FIELD_NUMBER: _ClassVar[int]
    LOGIT_BIAS_FIELD_NUMBER: _ClassVar[int]
    MIN_TOKENS_FIELD_NUMBER: _ClassVar[int]
    MAX_TOKENS_FIELD_NUMBER: _ClassVar[int]
    N_FIELD_NUMBER: _ClassVar[int]
    PRESENCE_PENALTY_FIELD_NUMBER: _ClassVar[int]
    STOP_FIELD_NUMBER: _ClassVar[int]
    STREAM_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_FIELD_NUMBER: _ClassVar[int]
    TOP_P_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_MICROSECONDS_FIELD_NUMBER: _ClassVar[int]
    LOGPROBS_FIELD_NUMBER: _ClassVar[int]
    TOP_LOGPROBS_FIELD_NUMBER: _ClassVar[int]
    TOP_K_FIELD_NUMBER: _ClassVar[int]
    REPETITION_PENALTY_FIELD_NUMBER: _ClassVar[int]
    SEED_FIELD_NUMBER: _ClassVar[int]
    EOS_TOKEN_FIELD_NUMBER: _ClassVar[int]
    TOOLS_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FORMAT_FIELD_NUMBER: _ClassVar[int]
    TOOL_CHOICE_FIELD_NUMBER: _ClassVar[int]
    PARALLEL_TOOL_CALLS_FIELD_NUMBER: _ClassVar[int]
    messages: _containers.RepeatedCompositeFieldContainer[Message]
    model: str
    frequency_penalty: float
    logit_bias: _containers.ScalarMap[int, float]
    min_tokens: int
    max_tokens: int
    n: int
    presence_penalty: float
    stop: _containers.RepeatedScalarFieldContainer[str]
    stream: bool
    temperature: float
    top_p: float
    timeout_microseconds: int
    logprobs: bool
    top_logprobs: int
    top_k: int
    repetition_penalty: float
    seed: _containers.RepeatedScalarFieldContainer[int]
    eos_token: _containers.RepeatedScalarFieldContainer[int]
    tools: _containers.RepeatedCompositeFieldContainer[Tool]
    response_format: _response_format_pb2.ResponseFormat
    tool_choice: _struct_pb2.Value
    parallel_tool_calls: bool
    def __init__(
        self,
        messages: _Optional[_Iterable[_Union[Message, _Mapping]]] = ...,
        model: _Optional[str] = ...,
        frequency_penalty: _Optional[float] = ...,
        logit_bias: _Optional[_Mapping[int, float]] = ...,
        min_tokens: _Optional[int] = ...,
        max_tokens: _Optional[int] = ...,
        n: _Optional[int] = ...,
        presence_penalty: _Optional[float] = ...,
        stop: _Optional[_Iterable[str]] = ...,
        stream: bool = ...,
        temperature: _Optional[float] = ...,
        top_p: _Optional[float] = ...,
        timeout_microseconds: _Optional[int] = ...,
        logprobs: bool = ...,
        top_logprobs: _Optional[int] = ...,
        top_k: _Optional[int] = ...,
        repetition_penalty: _Optional[float] = ...,
        seed: _Optional[_Iterable[int]] = ...,
        eos_token: _Optional[_Iterable[int]] = ...,
        tools: _Optional[_Iterable[_Union[Tool, _Mapping]]] = ...,
        response_format: _Optional[
            _Union[_response_format_pb2.ResponseFormat, _Mapping]
        ] = ...,
        tool_choice: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...,
        parallel_tool_calls: bool = ...,
    ) -> None: ...



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: completions.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from __future__ import annotations

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from friendli.schema.api.v1.codegen import response_format_pb2 as response__format__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x11\x63ompletions.proto\x12\x04orca\x1a\x15response_format.proto"\xad\x0e\n\x14V1CompletionsRequest\x12\x13\n\x06stream\x18\x01 \x01(\x08H\x00\x88\x01\x01\x12\x12\n\x05model\x18\x39 \x01(\tH\x01\x88\x01\x01\x12\x13\n\x06prompt\x18\x03 \x01(\tH\x02\x88\x01\x01\x12\x0e\n\x06tokens\x18\x04 \x03(\x05\x12!\n\x14timeout_microseconds\x18\x05 \x01(\x05H\x03\x88\x01\x01\x12\x17\n\nmax_tokens\x18\x06 \x01(\x05H\x04\x88\x01\x01\x12\x1d\n\x10max_total_tokens\x18\x07 \x01(\x05H\x05\x88\x01\x01\x12\x17\n\nmin_tokens\x18\x08 \x01(\x05H\x06\x88\x01\x01\x12\x1d\n\x10min_total_tokens\x18\t \x01(\x05H\x07\x88\x01\x01\x12\x0e\n\x01n\x18\n \x01(\x05H\x08\x88\x01\x01\x12\x16\n\tnum_beams\x18\x0b \x01(\x05H\t\x88\x01\x01\x12\x1b\n\x0elength_penalty\x18\x0c \x01(\x02H\n\x88\x01\x01\x12\x1b\n\x0e\x65\x61rly_stopping\x18\x0f \x01(\x08H\x0b\x88\x01\x01\x12\x1c\n\x0fno_repeat_ngram\x18\x11 \x01(\x05H\x0c\x88\x01\x01\x12$\n\x17\x65ncoder_no_repeat_ngram\x18\x12 \x01(\x05H\r\x88\x01\x01\x12\x1f\n\x12repetition_penalty\x18\x13 \x01(\x02H\x0e\x88\x01\x01\x12\'\n\x1a\x65ncoder_repetition_penalty\x18" \x01(\x02H\x0f\x88\x01\x01\x12\x1e\n\x11\x66requency_penalty\x18\x35 \x01(\x02H\x10\x88\x01\x01\x12\x1d\n\x10presence_penalty\x18\x36 \x01(\x02H\x11\x88\x01\x01\x12\x18\n\x0btemperature\x18\x14 \x01(\x02H\x12\x88\x01\x01\x12\x12\n\x05top_k\x18\x15 \x01(\x05H\x13\x88\x01\x01\x12\x12\n\x05top_p\x18\x16 \x01(\x02H\x14\x88\x01\x01\x12\x0c\n\x04stop\x18\x17 \x03(\t\x12=\n\x0bstop_tokens\x18\x18 \x03(\x0b\x32(.orca.V1CompletionsRequest.TokenSequence\x12\x0c\n\x04seed\x18\x1a \x03(\x04\x12\x1e\n\x16token_index_to_replace\x18\x1b \x03(\x05\x12\x1c\n\x14\x65mbedding_to_replace\x18\x1c \x03(\x02\x12H\n\x10\x62\x65\x61m_search_type\x18\x1d \x01(\x0e\x32).orca.V1CompletionsRequest.BeamSearchTypeH\x15\x88\x01\x01\x12*\n\x1d\x62\x65\x61m_compat_pre_normalization\x18\x1e \x01(\x08H\x16\x88\x01\x01\x12.\n!beam_compat_no_post_normalization\x18\x1f \x01(\x08H\x17\x88\x01\x01\x12\x11\n\tbad_words\x18  \x03(\t\x12\x41\n\x0f\x62\x61\x64_word_tokens\x18! \x03(\x0b\x32(.orca.V1CompletionsRequest.TokenSequence\x12"\n\x15include_output_logits\x18/ \x01(\x08H\x18\x88\x01\x01\x12$\n\x17include_output_logprobs\x18\x32 \x01(\x08H\x19\x88\x01\x01\x12\x1c\n\x14\x66orced_output_tokens\x18\x33 \x03(\x05\x12\x11\n\teos_token\x18. \x03(\x05\x12\x32\n\x0fresponse_format\x18= \x01(\x0b\x32\x14.orca.ResponseFormatH\x1a\x88\x01\x01\x1a\x1f\n\rTokenSequence\x12\x0e\n\x06tokens\x18\x01 \x03(\x05"G\n\x0e\x42\x65\x61mSearchType\x12\x11\n\rDETERMINISTIC\x10\x00\x12\x0e\n\nSTOCHASTIC\x10\x01\x12\x12\n\x0eNAIVE_SAMPLING\x10\x02\x42\t\n\x07_streamB\x08\n\x06_modelB\t\n\x07_promptB\x17\n\x15_timeout_microsecondsB\r\n\x0b_max_tokensB\x13\n\x11_max_total_tokensB\r\n\x0b_min_tokensB\x13\n\x11_min_total_tokensB\x04\n\x02_nB\x0c\n\n_num_beamsB\x11\n\x0f_length_penaltyB\x11\n\x0f_early_stoppingB\x12\n\x10_no_repeat_ngramB\x1a\n\x18_encoder_no_repeat_ngramB\x15\n\x13_repetition_penaltyB\x1d\n\x1b_encoder_repetition_penaltyB\x14\n\x12_frequency_penaltyB\x13\n\x11_presence_penaltyB\x0e\n\x0c_temperatureB\x08\n\x06_top_kB\x08\n\x06_top_pB\x13\n\x11_beam_search_typeB \n\x1e_beam_compat_pre_normalizationB$\n"_beam_compat_no_post_normalizationB\x18\n\x16_include_output_logitsB\x1a\n\x18_include_output_logprobsB\x12\n\x10_response_format"\x9e\x01\n\x15V1CompletionsResponse\x12\x30\n\x05\x65vent\x18\x01 \x01(\x0e\x32!.orca.V1CompletionsResponse.Event\x12\r\n\x05token\x18\x02 \x03(\x05\x12\x11\n\x04text\x18\x03 \x01(\tH\x00\x88\x01\x01"(\n\x05\x45vent\x12\x11\n\rTOKEN_SAMPLED\x10\x00\x12\x0c\n\x08\x43OMPLETE\x10\x01\x42\x07\n\x05_text2`\n\x15TextGenerationService\x12G\n\x08Generate\x12\x1a.orca.V1CompletionsRequest\x1a\x1b.orca.V1CompletionsResponse"\x00\x30\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "completions_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_V1COMPLETIONSREQUEST"]._serialized_start = 51
    _globals["_V1COMPLETIONSREQUEST"]._serialized_end = 1888
    _globals["_V1COMPLETIONSREQUEST_TOKENSEQUENCE"]._serialized_start = 1249
    _globals["_V1COMPLETIONSREQUEST_TOKENSEQUENCE"]._serialized_end = 1280
    _globals["_V1COMPLETIONSREQUEST_BEAMSEARCHTYPE"]._serialized_start = 1282
    _globals["_V1COMPLETIONSREQUEST_BEAMSEARCHTYPE"]._serialized_end = 1353
    _globals["_V1COMPLETIONSRESPONSE"]._serialized_start = 1891
    _globals["_V1COMPLETIONSRESPONSE"]._serialized_end = 2049
    _globals["_V1COMPLETIONSRESPONSE_EVENT"]._serialized_start = 2000
    _globals["_V1COMPLETIONSRESPONSE_EVENT"]._serialized_end = 2040
    _globals["_TEXTGENERATIONSERVICE"]._serialized_start = 2051
    _globals["_TEXTGENERATIONSERVICE"]._serialized_end = 2147
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Friendli Model SDK."""

# pylint: disable=arguments-differ

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Type

from friendli.client.graphql.model import ModelGqlClient
from friendli.schema.resource.v1.model import Model
from friendli.sdk.resource.base import ResourceApi


@dataclass
class ModelApi(ResourceApi[ModelGqlClient, Model, str]):
    """Model resource API."""

    client: ModelGqlClient

    @property
    def _resource_model(self) -> Type[Model]:
        return Model

    def create(self, *args, **kwargs) -> Model:
        """Create model."""
        raise NotImplementedError

    def get(self, eid: str, *args, **kwargs) -> Model:
        """Get model information."""
        raise NotImplementedError

    def list(self) -> List[Model]:
        """List models."""
        data = self.client.list(project_id=self._get_project_id())
        models = self._model_parse(data)
        return models



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli SDK errors."""

from __future__ import annotations

from typing import Any, List, Optional


class FriendliError(Exception):
    """Friendli exception base."""


class AuthenticationError(FriendliError):
    """Authentication failure error."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize AuthTokenNotFoundError."""
        super().__init__(f"Failed to authenticate: {detail}")


class AuthorizationError(FriendliError):
    """Not authorized to access."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize AuthorizationError."""
        super().__init__(f"Not authorized to access: {detail}")


class AuthTokenNotFoundError(FriendliError):
    """Auth token is not found."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize AuthTokenNotFoundError."""
        super().__init__(f"Auth token is not found: {detail}")


class LowServicePlanError(FriendliError):
    """The feature is not supported in your service plan."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize LowServicePlanError."""
        super().__init__(f"The feature is not supported in your service plan: {detail}")


class InvalidConfigError(FriendliError):
    """Invalid configuration provided."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize InvalidConfigError."""
        super().__init__(f"Invalid configuration provided: {detail}")


class InvalidAttributesError(FriendliError):
    """Checkpoint attributes are not valid."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize InvalidAttributesError."""
        super().__init__(f"Invalid checkpoint attributes: {detail}")


class InvalidPathError(FriendliError):
    """Invalid path provided."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize InvalidPathError."""
        super().__init__(f"Invalid path provided: {detail}")


class EntityTooLargeError(FriendliError):
    """Entity is too large."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize EntityTooLargeError."""
        super().__init__(f"Entity is too large: {detail}")


class NotFoundError(FriendliError):
    """Requested resource is not found."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize NotFoundError."""
        super().__init__(f"The resource is not found: {detail}")


class CheckpointConversionError(FriendliError):
    """Checkpoint conversion failure error."""

    def __init__(self, msg: str) -> None:
        """Initialize CheckpointConversionError."""
        super().__init__(f"Conversion failed: {msg}")


class NotSupportedCheckpointError(CheckpointConversionError):
    """Checkpoint is not supported."""

    def __init__(self, invalid_option: str, valid_options: List[Any]) -> None:
        """Initialize NotSupportedCheckpointError."""
        super().__init__(
            f"{invalid_option} is not supported. Please use one of {valid_options}."
        )


class QuantizationError(FriendliError):
    """Quantization failure error."""

    def __init__(self, msg: str) -> None:
        """Initialize QuantizationError."""
        super().__init__(f"Quantization failed. {msg}")


class NotSupportedQuantModeError(QuantizationError):
    """Quantization mode is not supported."""

    def __init__(self, invalid_option: str, valid_options: List[Any]) -> None:
        """Initialize NotSupportedQuantModeError."""
        super().__init__(
            f"{invalid_option} is not supported. Please use one of {valid_options}."
        )


class NotSupportedQuantConfigError(QuantizationError):
    """Quantization configuration is not supported."""

    def __init__(self, invalid_option: str, valid_options: List[str]) -> None:
        """Initialize NotSupportedQuantConfigError."""
        super().__init__(
            f"{invalid_option} is not supported. Please use {valid_options}."
        )


class TokenizerNotFoundError(FriendliError):
    """Cannot find Friendli-compatible tokenizer info."""

    def __init__(self, msg: Optional[str] = None) -> None:
        """Initialize TokenizerNotFoundError."""
        docs_link = (
            "https://huggingface.co/docs/transformers/main_classes/tokenizer#tokenizer"
        )
        default_msg = (
            "Friendli only supports Hugging Face 'fast' tokenizer. "
            f"Refer to {docs_link} to get more info."
        )
        super().__init__(msg or default_msg)


class RequestTimeoutError(FriendliError):
    """Request timeout error."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize NotFoundError."""
        super().__init__(f"Request timeout: {detail}")


class TransferError(FriendliError):
    """Object transfer failed."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize TransferError."""
        super().__init__(f"Transfer failed: {detail}")


class APIError(FriendliError):
    """Friendli API error."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize APIError."""
        super().__init__(f"API error: {detail}")


class SessionClosedError(FriendliError):
    """API Session is closed."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize SessionClosedError."""
        super().__init__(
            f"Cannot send requests to the API server because session is not opened: {detail}"
        )


class MaxRetriesExceededError(FriendliError):
    """Max retries exceeded."""

    def __init__(self, exc: Optional[Exception] = None) -> None:
        """Initialize MaxRetriesExceededError."""
        super().__init__(f"Max retries limit exceeded due to an error: {str(exc)}")


class NotSupportedError(FriendliError):
    """Feature is not supported."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize NotSupportedError."""
        super().__init__(f"The feature is not supported: {detail}")


class FriendliInternalError(FriendliError):
    """Internal error of Friendli."""

    def __init__(self, message: Optional[str] = None) -> None:
        """Initialize FriendliInternalError."""
        if message is None:
            message = "Please contact to FriendliAI"
        else:
            message = message.rstrip() + "\nPlease contant to FriendliAI"

        super().__init__(message)


class InvalidGenerationError(FriendliInternalError):
    """Invalid generation error."""

    def __init__(self, message: Optional[str] = None) -> None:
        """Initialize invalid generation error."""
        super().__init__(f"Invalid generation result: {message}")



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Friendli V1 Chat Completion Chunk Serving API Schemas."""

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel
from typing_extensions import Literal

from friendli.schema.api.v1.chat.completions import ChatCompletionTokenLogprob


class ChoiceDeltaFunctionCall(BaseModel):
    arguments: Optional[str] = None
    name: Optional[str] = None


class ChoiceDeltaToolCallFunction(BaseModel):
    arguments: Optional[str] = None
    name: Optional[str] = None


class ChoiceDeltaToolCall(BaseModel):
    index: int
    id: Optional[str] = None
    function: Optional[ChoiceDeltaToolCallFunction] = None
    type: Optional[Literal["function"]] = None


class ChoiceDelta(BaseModel):
    content: Optional[str] = None
    function_call: Optional[ChoiceDeltaFunctionCall] = None
    role: Optional[Literal["system", "user", "assistant", "tool"]] = None
    tool_calls: Optional[List[ChoiceDeltaToolCall]] = None


class ChoiceLogprobs(BaseModel):
    content: Optional[List[ChatCompletionTokenLogprob]] = None


class Choice(BaseModel):
    delta: ChoiceDelta
    finish_reason: Optional[Literal["stop", "length", "tool_calls"]] = None
    index: int
    logprobs: Optional[ChoiceLogprobs] = None


class ChatCompletionChunk(BaseModel):
    choices: List[Choice]
    created: int



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

"""Friendli V1 Chat Completion Serving API Schemas."""

from __future__ import annotations

from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel
from typing_extensions import Required, TypedDict


class MessageParam(TypedDict, total=False):
    """Message param schema."""

    role: Required[str]
    content: Required[str]


class ToolFunctionParam(TypedDict, total=False):
    """Tool function param schema."""

    name: Required[str]
    description: Optional[str]
    parameters: Dict[str, Any]


class ToolParam(TypedDict, total=False):
    """Tool param schema."""

    type: Required[str]
    function: Required[ToolFunctionParam]


class ResponseFormatParam(TypedDict, total=True):
    """Response format param schema."""

    type: Required[str]
    schema: Optional[str]


class Function(BaseModel):
    """Function schema."""

    arguments: str
    name: str


class ChatCompletionMessageToolCall(BaseModel):
    """Tool call schema."""

    id: str
    function: Function
    type: Literal["function"]


class Message(BaseModel):
    """Message schema."""

    role: Optional[str] = None
    content: Optional[str] = None
    tool_calls: Optional[List[ChatCompletionMessageToolCall]] = None


class TopLogprob(BaseModel):
    """Top logprob schema."""

    token: str
    bytes: Optional[List[int]] = None
    logprob: float


class ChatCompletionTokenLogprob(BaseModel):
    """Chat completion token log prob schema."""

    token: str
    bytes: Optional[List[int]] = None
    logprob: float
    top_logprobs: List[TopLogprob]


class ChoiceLogprobs(BaseModel):
    """Schema of log prob info."""

    content: Optional[List[ChatCompletionTokenLogprob]] = None


class ChatCompletionChoice(BaseModel):
    """Chat completion choice schema."""

    index: int
    message: Message
    finish_reason: str
    logprobs: Optional[ChoiceLogprobs] = None


class ChatCompletionUsage(BaseModel):
    """Chat completion usage schema."""

    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class ChatCompletion(BaseModel):
    """Chat completion schema."""

    choices: List[ChatCompletionChoice]
    usage: ChatCompletionUsage
    created: int



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Image Schemas."""

from __future__ import annotations

from typing import List, Literal, Union

from pydantic import AnyHttpUrl, BaseModel, Field
from typing_extensions import Annotated, TypeAlias

ImageResponseFormatParam: TypeAlias = Union[str, Literal["url", "png", "jpeg", "raw"]]


class ImageDataUrl(BaseModel):
    """Image data in url format."""

    format: Literal["url"]
    seed: int
    url: AnyHttpUrl


class ImageDataB64(BaseModel):
    """Image data in png format."""

    format: Literal["png", "jpeg", "raw"]
    seed: int
    b64_json: str


_ImageData = Annotated[
    Union[ImageDataUrl, ImageDataB64], Field(..., discriminator="format")
]


class Image(BaseModel):
    """Image data."""

    data: List[_ImageData]



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chat_completions.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from __future__ import annotations

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2

from friendli.schema.api.v1.codegen import response_format_pb2 as response__format__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x16\x63hat_completions.proto\x12\x04orca\x1a\x1cgoogle/protobuf/struct.proto\x1a\x15response_format.proto"|\n\x08ToolCall\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12)\n\x08\x66unction\x18\x03 \x01(\x0b\x32\x17.orca.ToolCall.Function\x1a+\n\x08\x46unction\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x01(\t"\xa5\x01\n\x07Message\x12\x14\n\x07\x63ontent\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x0c\n\x04role\x18\x02 \x01(\t\x12\x11\n\x04name\x18\x03 \x01(\tH\x01\x88\x01\x01\x12\x19\n\x0ctool_call_id\x18\x04 \x01(\tH\x02\x88\x01\x01\x12"\n\ntool_calls\x18\x05 \x03(\x0b\x32\x0e.orca.ToolCallB\n\n\x08_contentB\x07\n\x05_nameB\x0f\n\r_tool_call_id"\xac\x01\n\x04Tool\x12\x0c\n\x04type\x18\x01 \x01(\t\x12%\n\x08\x66unction\x18\x02 \x01(\x0b\x32\x13.orca.Tool.Function\x1ao\n\x08\x46unction\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x18\n\x0b\x64\x65scription\x18\x02 \x01(\tH\x00\x88\x01\x01\x12+\n\nparameters\x18\x03 \x01(\x0b\x32\x17.google.protobuf.StructB\x0e\n\x0c_description"\xf6\x07\n\x18V1ChatCompletionsRequest\x12\x1f\n\x08messages\x18\x01 \x03(\x0b\x32\r.orca.Message\x12\x12\n\x05model\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x1e\n\x11\x66requency_penalty\x18\x03 \x01(\x02H\x01\x88\x01\x01\x12\x41\n\nlogit_bias\x18\x04 \x03(\x0b\x32-.orca.V1ChatCompletionsRequest.LogitBiasEntry\x12\x17\n\nmin_tokens\x18\x05 \x01(\x05H\x02\x88\x01\x01\x12\x17\n\nmax_tokens\x18\x06 \x01(\x05H\x03\x88\x01\x01\x12\x0e\n\x01n\x18\x07 \x01(\x05H\x04\x88\x01\x01\x12\x1d\n\x10presence_penalty\x18\x08 \x01(\x02H\x05\x88\x01\x01\x12\x0c\n\x04stop\x18\t \x03(\t\x12\x13\n\x06stream\x18\n \x01(\x08H\x06\x88\x01\x01\x12\x18\n\x0btemperature\x18\x0b \x01(\x02H\x07\x88\x01\x01\x12\x12\n\x05top_p\x18\x0c \x01(\x02H\x08\x88\x01\x01\x12!\n\x14timeout_microseconds\x18\r \x01(\x05H\t\x88\x01\x01\x12\x15\n\x08logprobs\x18\x0e \x01(\x08H\n\x88\x01\x01\x12\x19\n\x0ctop_logprobs\x18\x0f \x01(\x05H\x0b\x88\x01\x01\x12\x12\n\x05top_k\x18\x13 \x01(\x05H\x0c\x88\x01\x01\x12\x1f\n\x12repetition_penalty\x18\x14 \x01(\x02H\r\x88\x01\x01\x12\x0c\n\x04seed\x18\x15 \x03(\x04\x12\x11\n\teos_token\x18\x16 \x03(\x05\x12\x19\n\x05tools\x18\x17 \x03(\x0b\x32\n.orca.Tool\x12\x32\n\x0fresponse_format\x18\x18 \x01(\x0b\x32\x14.orca.ResponseFormatH\x0e\x88\x01\x01\x12\x30\n\x0btool_choice\x18\x19 \x01(\x0b\x32\x16.google.protobuf.ValueH\x0f\x88\x01\x01\x12 \n\x13parallel_tool_calls\x18\x1a \x01(\x08H\x10\x88\x01\x01\x1a\x30\n\x0eLogitBiasEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x42\x08\n\x06_modelB\x14\n\x12_frequency_penaltyB\r\n\x0b_min_tokensB\r\n\x0b_max_tokensB\x04\n\x02_nB\x13\n\x11_presence_penaltyB\t\n\x07_streamB\x0e\n\x0c_temperatureB\x08\n\x06_top_pB\x17\n\x15_timeout_microsecondsB\x0b\n\t_logprobsB\x0f\n\r_top_logprobsB\x08\n\x06_top_kB\x15\n\x13_repetition_penaltyB\x12\n\x10_response_formatB\x0e\n\x0c_tool_choiceB\x16\n\x14_parallel_tool_callsb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chat_completions_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_V1CHATCOMPLETIONSREQUEST_LOGITBIASENTRY"]._loaded_options = None
    _globals["_V1CHATCOMPLETIONSREQUEST_LOGITBIASENTRY"]._serialized_options = b"8\001"
    _globals["_TOOLCALL"]._serialized_start = 85
    _globals["_TOOLCALL"]._serialized_end = 209
    _globals["_TOOLCALL_FUNCTION"]._serialized_start = 166
    _globals["_TOOLCALL_FUNCTION"]._serialized_end = 209
    _globals["_MESSAGE"]._serialized_start = 212
    _globals["_MESSAGE"]._serialized_end = 377
    _globals["_TOOL"]._serialized_start = 380
    _globals["_TOOL"]._serialized_end = 552
    _globals["_TOOL_FUNCTION"]._serialized_start = 441
    _globals["_TOOL_FUNCTION"]._serialized_end = 552
    _globals["_V1CHATCOMPLETIONSREQUEST"]._serialized_start = 555
    _globals["_V1CHATCOMPLETIONSREQUEST"]._serialized_end = 1569
    _globals["_V1CHATCOMPLETIONSREQUEST_LOGITBIASENTRY"]._serialized_start = 1247
    _globals["_V1CHATCOMPLETIONSREQUEST_LOGITBIASENTRY"]._serialized_end = 1295
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Client Service."""

from __future__ import annotations

import copy
import math
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import wraps
from pathlib import Path
from string import Template
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar, Union
from urllib.parse import urljoin, urlparse

import requests
from requests import HTTPError
from requests.models import Response
from urllib3.exceptions import ConnectTimeoutError, ReadTimeoutError

from friendli.auth import get_auth_header, safe_request
from friendli.context import get_current_project_id, get_current_team_id
from friendli.di.injector import get_injector
from friendli.errors import APIError, MaxRetriesExceededError
from friendli.logging import logger
from friendli.utils.format import secho_error_and_exit
from friendli.utils.fs import get_file_size
from friendli.utils.request import DEFAULT_REQ_TIMEOUT
from friendli.utils.transfer import S3_MULTIPART_CHUNK_SIZE, ChunksizeAdjuster
from friendli.utils.url import URLProvider

T = TypeVar("T", bound=Union[int, str, uuid.UUID])

RETRYABLE_REQUEST_ERRORS = (
    ConnectionError,
    requests.exceptions.ReadTimeout,
    requests.exceptions.ConnectionError,
    ReadTimeoutError,
    ConnectTimeoutError,
)
API_REQUEST_MAX_RETRIES = 3
DEFAULT_PAGINATION_SIZE = 20


@dataclass
class URLTemplate:
    """URL template."""

    pattern: Template

    def render(
        self, pk: Optional[T] = None, path: Optional[str] = None, **kwargs
    ) -> str:
        """Render URLTemplate.

        Args:
            pk: Primary key of a resource.
            path: Additional URL path to attach.

        """
        if pk is None and path is None:
            return self.pattern.substitute(**kwargs)

        pattern = copy.deepcopy(self.pattern)
        need_trailing_slash = pattern.template.endswith("/")

        if pk is not None:
            pattern.template = urljoin(pattern.template + "/", str(pk))
            if need_trailing_slash:
                pattern.template += "/"

        if path is not None:
            pattern.template = urljoin(pattern.template + "/", path.rstrip("/"))
            if need_trailing_slash:
                pattern.template += "/"

        return pattern.substitute(**kwargs)

    def get_base_url(self) -> str:
        """Get a base URL."""
        result = urlparse(self.pattern.template)
        return f"{result.scheme}://{result.hostname}"

    def attach_pattern(self, pattern: str) -> None:
        """Attach a URL path pattern."""
        self.pattern.template = urljoin(self.pattern.template + "/", pattern)

    def replace_path(self, path: str):
        """Replace a URL path."""
        result = urlparse(self.pattern.template)
        result = result._replace(path=path)
        self.pattern.template = result.geturl()

    def copy(self) -> "URLTemplate":
        """Get a copy of this URL template."""
        return URLTemplate(pattern=Template(self.pattern.template))


class RequestInterface:
    """Request API mixin."""

    def check_request(self, func: Callable[..., Response]) -> Callable[..., Response]:
        """Wrapper function to send requests with the retry and error handling."""

        @wraps(func)
        def wrapper(*args, **kwargs) -> Response:  # type: ignore
            final_exc = None
            for i in range(API_REQUEST_MAX_RETRIES):
                try:
                    return func(*args, **kwargs)
                except RETRYABLE_REQUEST_ERRORS:
                    logger.info(
                        ("Retry the failed request (attempt %s / %s)."),
                        i + 1,
                        API_REQUEST_MAX_RETRIES,
                    )
                except Exception as exc:  # pylint: disable=broad-exception-caught
                    final_exc = exc
            raise MaxRetriesExceededError(final_exc)

        return wrapper

    def paginated_get(
        self,
        callback: Callable[..., Response],
        path: Optional[str] = None,
        limit: int = DEFAULT_PAGINATION_SIZE,
        params: Optional[dict[str, Any]] = None,
    ) -> List[Dict[str, Any]]:
        """List objects with pagination."""
        page_size = min(DEFAULT_PAGINATION_SIZE, limit)
        params = {"limit": page_size, **(params or {})}
        response_dict = callback(path=path, params=params).json()
        items = response_dict["results"]
        next_cursor = response_dict["next_cursor"]

        while next_cursor is not None and len(items) < limit:
            response_dict = callback(
                path=path, params={**params, "cursor": next_cursor}
            ).json()
            items.extend(response_dict["results"])
            next_cursor = response_dict["next_cursor"]

        return items


class HttpClient(ABC, Generic[T], RequestInterface):
    """Base interface of client to Friendli system."""

    def __init__(self, **kwargs):
        """Initialize client."""
        self.injector = get_injector()
        self.url_provider = self.injector.get(URLProvider)
        self.url_template = URLTemplate(Template(self.url_path))
        self.url_kwargs = kwargs

    @property
    @abstractmethod
    def url_path(self) -> str:
        """URL path template to render."""

    @property
    def default_request_options(self) -> Dict[str, Any]:
        """Common request options."""
        return {
            "headers": get_auth_header(),
            "timeout": DEFAULT_REQ_TIMEOUT,
        }

    def list(
        self,
        path: Optional[str] = None,
        *,
        pagination: bool = True,
        limit: int = DEFAULT_PAGINATION_SIZE,
        **kwargs,
    ) -> Any:
        """Send a request to list objects."""
        if pagination:
            page_size = min(DEFAULT_PAGINATION_SIZE, limit)
            params = kwargs.pop("params", {})
            params = {"limit": page_size, "page_size": page_size, **params}
            data = self._list(path=path, params=params, **kwargs)
            items = data["results"]
            next_cursor = data["next_cursor"]

            while next_cursor is not None and len(items) < limit:
                params["limit"] = params["page_size"] = min(
                    page_size, limit - len(items)
                )
                data = self._list(
                    path=path,
                    params={**params, "cursor": next_cursor},
                    **kwargs,
                )
                items.extend(data["results"])
                next_cursor = data["next_cursor"]

            return items
        return self._list(path=path, **kwargs)

    @safe_request
    def retrieve(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to retrieve a specific object."""
        resp = self.check_request(requests.get)(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )
        return resp

    @safe_request
    def post(self, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to post an object."""
        resp = self.check_request(requests.post)(
            self.url_template.render(path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )
        return resp

    @safe_request
    def partial_update(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to partially update a specific object."""
        resp = self.check_request(requests.patch)(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )
        return resp

    @safe_request
    def delete(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to delete a specific obejct."""
        resp = self.check_request(requests.delete)(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )
        return resp

    @safe_request
    def update(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to update a specific object."""
        resp = self.check_request(requests.put)(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )
        return resp

    def bare_post(self, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request without automatic auth."""
        resp = self.check_request(requests.post)(
            self.url_template.render(path=path, **self.url_kwargs),
            timeout=DEFAULT_REQ_TIMEOUT,
            **kwargs,
        )
        try:
            resp.raise_for_status()
        except HTTPError as exc:
            raise APIError(str(exc)) from exc
        return resp

    @safe_request
    def _list(self, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to list objects."""
        resp = self.check_request(requests.get)(
            self.url_template.render(path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )
        return resp


class UserRequestMixin(RequestInterface):
    """Mixin class of user-level requests."""

    user_id: uuid.UUID

    @safe_request
    def get_current_user_info(self) -> Response:
        """Gets the current user info."""
        injector = get_injector()
        url_provider = injector.get(URLProvider)
        return self.check_request(requests.get)(
            url_provider.get_auth_uri("oauth2/userinfo"),
            headers=get_auth_header(),
            timeout=DEFAULT_REQ_TIMEOUT,
        )

    def get_current_user_id(self) -> uuid.UUID:
        """Gets the currently logged-in user ID."""
        userinfo = self.get_current_user_info()
        return uuid.UUID(userinfo["sub"].split("|")[1])

    def initialize_user(self):
        """Initializes user settings."""
        self.user_id = self.get_current_user_id()


class GroupRequestMixin:
    """Mixin class of organization-level requests."""

    group_id: uuid.UUID

    def initialize_group(self):
        """Initialize organization settings."""
        group_id = get_current_team_id()
        if group_id is None:
            secho_error_and_exit("Organization is not set.")
        self.group_id = group_id  # type: ignore


class ProjectRequestMixin:
    """Mixin class of project-level requests."""

    project_id: uuid.UUID

    def initialize_project(self):
        """Initialize project settings."""
        project_id = get_current_project_id()
        if project_id is None:
            secho_error_and_exit("Project is not set.")
        self.project_id = project_id  # type: ignore


class UploadableClient(HttpClient[T], Generic[T]):
    """Uploadable client."""

    def get_upload_urls(
        self,
        obj_id: T,
        storage_paths: List[str],
    ) -> List[Dict[str, Any]]:
        """Get single part upload URLs for multiple files.

        Args:
            obj_id (T): Uploadable object ID
            storage_paths (List[str]): A list of cloud storage paths of target files

        Returns:
            List[Dict[str, Any]]: A response body that has presigned URL info to upload files.

        """
        data = self.post(path=f"{obj_id}/upload/", json={"paths": storage_paths})
        return data

    def get_multipart_upload_urls(
        self,
        obj_id: T,
        local_paths: List[Path],
        storage_paths: List[str],
    ) -> List[Dict[str, Any]]:
        """Get multipart upload URLs for multiple file-like objects.

        Args:
            obj_id (T): Uploadable object ID
            local_paths (List[Path]): A list local paths to target files. The path can
                be either absolute or relative.
            storage_paths (List[str]): A list of storage paths to target files.

        Returns:
            List[Dict[str, Any]]: A list of multipart upload responses for multiple target files.

        """
        start_mpu_resps = []
        adjuster = ChunksizeAdjuster()
        for local_path, storage_path in zip(local_paths, storage_paths):
            file_size = get_file_size(local_path)
            part_size = adjuster.adjust_chunksize(
                current_chunksize=S3_MULTIPART_CHUNK_SIZE, file_size=file_size
            )
            num_parts = math.ceil(file_size / part_size)
            data = self.post(
                path=f"{obj_id}/start_mpu/",
                json={
                    "path": storage_path,
                    "num_parts": num_parts,
                },
            )
            start_mpu_resps.append(data)
        return start_mpu_resps

    def complete_multipart_upload(
        self, obj_id: T, path: str, upload_id: str, parts: List[Dict[str, Any]]
    ) -> None:
        """Complete multipart upload.

        Args:
            obj_id (T): Uploadable object ID
            path (str): Path to the uploaded file
            upload_id (str): Upload ID
            parts (List[Dict[str, Any]]): A list of upload part info

        """
        sorted_parts = sorted(parts, key=lambda part: part["part_number"])
        self.post(
            path=f"{obj_id}/complete_mpu/",
            json={
                "path": path,
                "upload_id": upload_id,
                "parts": sorted_parts,
            },
        )

    def abort_multipart_upload(self, obj_id: T, path: str, upload_id: str) -> None:
        """Abort multipart upload.

        Args:
            obj_id (T): Uploadable object ID
            path (str): Path to the target file
            upload_id (str): Upload ID
            parts (List[Dict[str, Any]]): A list of upload part info

        """
        self.post(
            path=f"{obj_id}/abort_mpu/",
            json={
                "path": path,
                "upload_id": upload_id,
            },
        )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI."""

# pylint: disable=line-too-long

from __future__ import annotations

import typer

import friendli
from friendli.auth import TokenType, clear_tokens, update_token
from friendli.cli import api, endpoint, model, project, team
from friendli.cli.login import oauth2_login, pwd_login
from friendli.client.graphql.team import TeamGqlClient
from friendli.client.graphql.user import UserGqlClient
from friendli.context import (
    clear_contexts,
    get_current_project_id,
    get_current_team_id,
    set_current_project_id,
    set_current_team_id,
)
from friendli.errors import AuthTokenNotFoundError
from friendli.formatter import PanelFormatter
from friendli.utils.decorator import check_api
from friendli.utils.format import secho_error_and_exit
from friendli.utils.version import get_installed_version

app = typer.Typer(
    help="Supercharge Generative AI Serving 🚀",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=True,
    pretty_exceptions_enable=False,
)

app.add_typer(endpoint.app, name="endpoint", help="Manage endpoints")
app.add_typer(project.app, name="project", help="Manage projects")
app.add_typer(team.app, name="team", help="Manage teams")
app.add_typer(model.app, name="model", help="Manage models")
app.add_typer(api.app, name="api", help="API call to endpoints")


user_panel_formatter = PanelFormatter(
    name="My Info",
    fields=["name", "email"],
    headers=["Name", "Email"],
)


@app.command()
@check_api
def whoami():
    """Show my user info."""
    try:
        client = UserGqlClient()
        info = client.get_current_user_info()
    except AuthTokenNotFoundError as exc:
        secho_error_and_exit(str(exc))

    user_panel_formatter.render([info])


@app.command()
@check_api
def login(
    use_sso: bool = typer.Option(False, "--sso", help="Use SSO login."),
):
    """Sign in Friendli."""
    clear_tokens()
    clear_contexts()

    if friendli.token:
        typer.secho(
            "You've already set the 'FRIENDLI_TOKEN' environment variable for "
            "authentication, which takes precedence over the login session. Using both "
            "authentication methods could lead to unexpected issues. We suggest "
            "unsetting the 'FRIENDLI_TOKEN' environment variable if you prefer to log "
            "in through the standard login session.",
            fg=typer.colors.RED,
        )

    if use_sso:
        access_token, refresh_token = oauth2_login()
    else:
        email = typer.prompt("Enter your email")
        pwd = typer.prompt("Enter your password", hide_input=True)
        access_token, refresh_token = pwd_login(email, pwd)

    update_token(token_type=TokenType.ACCESS, token=access_token)
    update_token(token_type=TokenType.REFRESH, token=refresh_token)

    # Set default team and project.
    team_id = _set_default_team()
    if team_id is not None:
        _set_default_project(team_id)

    _display_login_success()


@app.command()
def logout():
    """Sign out."""
    clear_tokens()
    clear_contexts()
    typer.secho("Successfully signed out.", fg=typer.colors.BLUE)


@app.command()
def version():
    """Check the installed package version."""
    installed_version = get_installed_version()
    typer.echo(installed_version)


def _display_login_success():
    typography = r"""
,------.       ,--.                  ,--.,--.,--. 
|  .---',--.--.`--' ,---. ,--,--,  ,-|  ||  |`--' 
|  `--, |  .--',--.| .-. :|      \' .-. ||  |,--. 
|  |`   |  |   |  |\   --.|  ||  |\ `-' ||  ||  | 
`--'    `--'   `--' `----'`--''--' `---' `--'`--' 
"""
    typer.secho(f"\nLOGIN SUCCESS!\n{typography}", fg=typer.colors.BLUE)


def _set_default_team() -> str | None:
    current_team_id = get_current_team_id()
    if current_team_id is None:
        client = UserGqlClient()
        default_team_id = client.get_default_team()["id"]
        set_current_team_id(default_team_id)
        typer.secho(
            f"📌 CLI configures your team to '{default_team_id}'. If you want to "
            "switch the team, run 'friendli team switch'.",
            fg=typer.colors.YELLOW,
        )
        return default_team_id

    return current_team_id


def _set_default_project(team_id: str) -> str | None:
    current_project_id = get_current_project_id()
    if current_project_id is None:
        client = TeamGqlClient()
        projects = client.get_projects(team_id=team_id)
        if len(projects) > 0:
            default_project_id = projects[0]["node"]["id"]
            set_current_project_id(default_project_id)
            typer.secho(
                f"📌 CLI configures your project to '{default_project_id}'. If you want "
                "to switch the project, run 'friendli project switch'.",
                fg=typer.colors.YELLOW,
            )
            return default_project_id

        typer.secho(
            "Project authorized to access is not found. Contact to your team admin.",
            fg=typer.colors.RED,
        )
        return None

    return current_project_id



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Completion API."""

# pylint: disable=line-too-long, no-name-in-module, too-many-locals

from __future__ import annotations

import json
from typing import Dict, List, Literal, Optional, Type, Union, overload

from pydantic import ValidationError

from friendli.errors import InvalidGenerationError
from friendli.schema.api.v1.chat.completion_chunk import ChatCompletionChunk
from friendli.schema.api.v1.chat.completions import (
    ChatCompletion,
    MessageParam,
    ResponseFormatParam,
    ToolParam,
)
from friendli.schema.api.v1.codegen.chat_completions_pb2 import V1ChatCompletionsRequest
from friendli.sdk.api.base import (
    AsyncGenerationStream,
    AsyncServingAPI,
    GenerationStream,
    ServingAPI,
)
from friendli.utils.compat import model_parse


class Completions(ServingAPI[Type[V1ChatCompletionsRequest]]):
    """Friendli completions API."""

    @property
    def _api_path(self) -> str:
        return "v1/chat/completions"

    @property
    def _method(self) -> str:
        return "POST"

    @property
    def _content_type(self) -> str:
        return "application/protobuf" if self._use_protobuf else "application/json"

    @property
    def _request_pb_cls(self) -> Type[V1ChatCompletionsRequest]:
        return V1ChatCompletionsRequest

    @overload
    def create(
        self,
        *,
        messages: List[MessageParam],
        stream: Literal[True],
        model: Optional[str] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        logit_bias: Optional[Dict[int, float]] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        timeout_microseconds: Optional[int] = None,
        tools: Optional[List[ToolParam]] = None,
        tool_choice: Optional[str] = "auto",
        response_format: Optional[ResponseFormatParam] = None,
    ) -> ChatCompletionStream:
        """[skip-doc]."""

    @overload
    def create(
        self,
        *,
        messages: List[MessageParam],
        stream: Literal[False],
        model: Optional[str] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        logit_bias: Optional[Dict[int, float]] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        timeout_microseconds: Optional[int] = None,
        tools: Optional[List[ToolParam]] = None,
        tool_choice: Optional[str] = "auto",
        response_format: Optional[ResponseFormatParam] = None,
    ) -> ChatCompletion:
        """[skip-doc]."""

    def create(
        self,
        *,
        messages: List[MessageParam],
        stream: bool = False,
        model: Optional[str] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        logit_bias: Optional[Dict[int, float]] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        timeout_microseconds: Optional[int] = None,
        tools: Optional[List[ToolParam]] = None,
        tool_choice: Optional[str] = "auto",
        response_format: Optional[ResponseFormatParam] = None,
    ) -> Union[ChatCompletionStream, ChatCompletion]:
        """Creates a chat completion.

        Args:
            messages (List[MessageParam]): A list of messages comprising the conversation so far.
            stream (bool, optional): Whether to stream generation result. When set true, each token will be sent as server-sent events once generated. Not supported when using beam search.
            model (Optional[str]): Code of the model to use. This argument should be set only for serverless endpoints.
            frequency_penalty (Optional[float], optional): Number between -2.0 and 2.0. Positive values penalizes tokens that have been sampled, taking into account their frequency in the preceding text. This penalization diminishes the model's tendency to reproduce identical lines verbatim. Defaults to None.
            presence_penalty (Optional[float], optional): Number between -2.0 and 2.0. Positive values penalizes tokens that have been sampled at least once in the existing text. Defaults to None.
            max_tokens (Optional[int], optional): The maximum number of tokens to generate. The length of your input tokens plus `max_tokens` should not exceed the model's maximum length (e.g., 2048 for OpenAI GPT-3). Defaults to None.
            n (Optional[int], optional): The number of independently generated results for the prompt. Defaults to None.
            stop (Optional[List[str]], optional): When one of the stop phrases appears in the generation result, the API will stop generation. The phrase is included in the generated result. If you are using beam search, all of the active beams should contain the stop phrase to terminate generation. Before checking whether a stop phrase is included in the result, the phrase is converted into tokens. We recommend using `stop_tokens` because it is clearer. For example, after tokenization, phrases "clear" and " clear" can result in different token sequences due to the prepended space character. Defaults to None.
            temperature (Optional[float], optional): Sampling temperature. Non-zero positive numbers are allowed. Smaller temperature makes the generation result closer to greedy, argmax (i.e., `top_k = 1`) sampling. If it is `None`, then 1.0 is used by default. This is similar to Hugging Face's [`temperature`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.temperature) argument. Defaults to None.
            top_p (Optional[float], optional): Tokens comprising the top `top_p` probability mass are kept for sampling. Numbers between 0.0 (exclusive) and 1.0 (inclusive) are allowed. If it is `None`, then 1.0 is used by default. This is similar to Hugging Face's [`top_p`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.top_p) argument. Defaults to None.
            timeout_microseconds (Optional[int], optional): Request timeout. Gives the HTTP `429 Too Many Requests` response status code. Default behavior is no timeout. Defaults to None.

        Returns:
            Union[ChatCompletionStream, ChatCompletion]: If `stream` is `True`, then `ChatCompletionStream` object that iterates the results per token is returned. Otherwise, a `ChatCompletion` object is returend.

        """
        request_dict = {
            "messages": messages,
            "stream": stream,
            "model": model,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "repetition_penalty": repetition_penalty,
            "max_tokens": max_tokens,
            "min_tokens": min_tokens,
            "n": n,
            "stop": stop,
            "temperature": temperature,
            "top_p": top_p,
            "top_k": top_k,
            "logit_bias": logit_bias,
            "logprobs": logprobs,
            "top_logprobs": top_logprobs,
            "timeout_microseconds": timeout_microseconds,
            "tools": tools,
            "tool_choice": tool_choice,
            "response_format": response_format,
        }
        response = self._request(data=request_dict, stream=stream)

        if stream:
            return ChatCompletionStream(response=response)
        return model_parse(ChatCompletion, response.json())


class AsyncCompletions(AsyncServingAPI[Type[V1ChatCompletionsRequest]]):
    """Async completions."""

    @property
    def _api_path(self) -> str:
        return "v1/chat/completions"

    @property
    def _method(self) -> str:
        return "POST"

    @property
    def _content_type(self) -> str:
        return "application/protobuf" if self._use_protobuf else "application/json"

    @property
    def _request_pb_cls(self) -> Type[V1ChatCompletionsRequest]:
        return V1ChatCompletionsRequest

    @overload
    async def create(
        self,
        *,
        messages: List[MessageParam],
        stream: Literal[True],
        model: Optional[str] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        logit_bias: Optional[Dict[int, float]] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        timeout_microseconds: Optional[int] = None,
        tools: Optional[List[ToolParam]] = None,
        tool_choice: Optional[str] = "auto",
        response_format: Optional[ResponseFormatParam] = None,
    ) -> AsyncChatCompletionStream:
        """[skip-doc]."""

    @overload
    async def create(
        self,
        *,
        messages: List[MessageParam],
        stream: Literal[False],
        model: Optional[str] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        logit_bias: Optional[Dict[int, float]] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        timeout_microseconds: Optional[int] = None,
        tools: Optional[List[ToolParam]] = None,
        tool_choice: Optional[str] = "auto",
        response_format: Optional[ResponseFormatParam] = None,
    ) -> ChatCompletion:
        """[skip-doc]."""

    async def create(
        self,
        *,
        messages: List[MessageParam],
        stream: bool = False,
        model: Optional[str] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        logit_bias: Optional[Dict[int, float]] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        timeout_microseconds: Optional[int] = None,
        tools: Optional[List[ToolParam]] = None,
        tool_choice: Optional[str] = "auto",
        response_format: Optional[ResponseFormatParam] = None,
    ) -> Union[AsyncChatCompletionStream, ChatCompletion]:
        """Creates a completion asynchronously.

        Args:
            messages (List[MessageParam]): A list of messages comprising the conversation so far.
            stream (bool, optional): Whether to stream generation result. When set true, each token will be sent as server-sent events once generated. Not supported when using beam search.
            model (Optional[str]): Code of the model to use. This argument should be set only for serverless endpoints.
            frequency_penalty (Optional[float], optional): Number between -2.0 and 2.0. Positive values penalizes tokens that have been sampled, taking into account their frequency in the preceding text. This penalization diminishes the model's tendency to reproduce identical lines verbatim. Defaults to None.
            presence_penalty (Optional[float], optional): Number between -2.0 and 2.0. Positive values penalizes tokens that have been sampled at least once in the existing text. Defaults to None.
            max_tokens (Optional[int], optional): The maximum number of tokens to generate. The length of your input tokens plus `max_tokens` should not exceed the model's maximum length (e.g., 2048 for OpenAI GPT-3). Defaults to None.
            n (Optional[int], optional): The number of independently generated results for the prompt. Defaults to None.
            stop (Optional[List[str]], optional): When one of the stop phrases appears in the generation result, the API will stop generation. The phrase is included in the generated result. If you are using beam search, all of the active beams should contain the stop phrase to terminate generation. Before checking whether a stop phrase is included in the result, the phrase is converted into tokens. We recommend using `stop_tokens` because it is clearer. For example, after tokenization, phrases "clear" and " clear" can result in different token sequences due to the prepended space character. Defaults to None.
            temperature (Optional[float], optional): Sampling temperature. Non-zero positive numbers are allowed. Smaller temperature makes the generation result closer to greedy, argmax (i.e., `top_k = 1`) sampling. If it is `None`, then 1.0 is used by default. This is similar to Hugging Face's [`temperature`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.temperature) argument. Defaults to None.
            top_p (Optional[float], optional): Tokens comprising the top `top_p` probability mass are kept for sampling. Numbers between 0.0 (exclusive) and 1.0 (inclusive) are allowed. If it is `None`, then 1.0 is used by default. This is similar to Hugging Face's [`top_p`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.top_p) argument. Defaults to None.
            timeout_microseconds (Optional[int], optional): Request timeout. Gives the HTTP `429 Too Many Requests` response status code. Default behavior is no timeout. Defaults to None.

        Returns:
            Union[AsyncChatCompletionStream, ChatCompletion]: If `stream` is `True`, then `AsyncChatCompletionStream` object that iterates the results per token is returned. Otherwise, a `ChatCompletion` object is returend.

        """
        request_dict = {
            "messages": messages,
            "stream": stream,
            "model": model,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "repetition_penalty": repetition_penalty,
            "max_tokens": max_tokens,
            "min_tokens": min_tokens,
            "n": n,
            "stop": stop,
            "temperature": temperature,
            "top_p": top_p,
            "top_k": top_k,
            "logit_bias": logit_bias,
            "logprobs": logprobs,
            "top_logprobs": top_logprobs,
            "timeout_microseconds": timeout_microseconds,
            "tools": tools,
            "tool_choice": tool_choice,
            "response_format": response_format,
        }
        response = await self._request(data=request_dict, stream=stream)

        if stream:
            return AsyncChatCompletionStream(response=response)
        return model_parse(ChatCompletion, response.json())


class ChatCompletionStream(GenerationStream[ChatCompletionChunk]):
    """Completion stream."""

    def __next__(self) -> ChatCompletionChunk:  # noqa: D105
        line = next(self._iter)
        while not line:
            line = next(self._iter)

        data = line.strip("data: ")
        if data == "[DONE]":
            self._response.close()
            raise StopIteration
        parsed = json.loads(data)

        try:
            return model_parse(ChatCompletionChunk, parsed)
        except ValidationError as exc:
            self._response.close()
            raise InvalidGenerationError(
                f"Generation result has invalid schema: {str(exc)}"
            ) from exc


class AsyncChatCompletionStream(AsyncGenerationStream[ChatCompletionChunk]):
    """Asynchronous completion stream."""

    async def __anext__(self) -> ChatCompletionChunk:  # noqa: D105
        line = await self._iter.__anext__()
        while not line:
            line = await self._iter.__anext__()

        data = line.strip("data: ")
        if data == "[DONE]":
            await self._response.aclose()
            raise StopAsyncIteration
        parsed = json.loads(data)

        try:
            return model_parse(ChatCompletionChunk, parsed)
        except ValidationError as exc:
            await self._response.aclose()
            raise InvalidGenerationError(
                f"Generation result has invalid schema: {str(exc)}"
            ) from exc



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI Validation Utilities."""

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, Optional, Type

import typer
from pydantic import ValidationError

from friendli.errors import InvalidAttributesError, InvalidConfigError
from friendli.schema.resource.v1.attributes import V1AttributesValidationModel
from friendli.utils.compat import model_parse
from friendli.utils.version import (
    FRIENDLI_PACKAGE_NAME,
    get_installed_version,
    get_latest_version,
    is_latest_version,
)


def validate_datetime_format(datetime_str: Optional[str]) -> Optional[str]:
    """Validate the datetime format."""
    if datetime_str is None:
        return None

    try:
        datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S")
    except ValueError as exc:
        raise typer.BadParameter(
            "The datetime format should be {YYYY}-{MM}-{DD}T{HH}:{MM}:{SS}"
        ) from exc
    return datetime_str


def check_package_version() -> None:
    """Validate the installed CLI version."""
    installed_version = get_installed_version()
    if not is_latest_version(installed_version):
        latest_version = get_latest_version()
        typer.secho(
            f"Package version({installed_version}) is not the latest. "
            f"We recommend installing the latest version({latest_version}) with "
            f"'pip install {FRIENDLI_PACKAGE_NAME}=={latest_version} -U'.",
            fg=typer.colors.YELLOW,
        )


def validate_checkpoint_attributes(attr: Dict[str, Any]) -> None:
    """Validate checkpoint attributes schema."""
    try:
        model_parse(V1AttributesValidationModel, {"attr": attr})
    except ValidationError as exc:
        msgs = []
        for error in exc.errors():
            error_type = error["type"]
            if error_type == "union_tag_invalid":
                msgs.append("'model_type' is not one of the expected values.")
            elif error_type == "union_tag_not_found":
                msgs.append("'model_type' filed is missing.")
            else:
                msgs.append(f"{error['msg']}. Correct the field '{error['loc'][-1]}'")

        if len(msgs) == 1:
            msg = msgs[0]
        else:
            msg = "\n>>> " + "\n>>> ".join(msgs)
        raise InvalidAttributesError(msg) from exc


def validate_enums(val: Any, enum_cls: Type[Enum]) -> Any:
    """Validate if the value is the proper enum."""
    try:
        return enum_cls(val)
    except ValueError as exc:
        supported_values = set([e.value for e in enum_cls])
        raise InvalidConfigError(
            f"Invalid value. Please provide one of {supported_values}"
        ) from exc



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Endpoint Schemas."""

from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from friendli.schema.resource.v1.common import Creator


class EndpointPhase(BaseModel):
    """Endpoint phase schema."""

    step: Optional[str] = None
    msg: Optional[str] = None
    desiredReplica: Optional[int] = None
    currReplica: Optional[int] = None


class EndpointAdvancedConfig(BaseModel):
    """Endpoint advanced config schema."""

    maxBatchSize: Optional[int] = None
    autoscalingMin: Optional[int] = None
    autoscalingMax: Optional[int] = None


class Endpoint(BaseModel):
    """Endpoint schema."""

    id: Optional[str]
    name: Optional[str]
    hfModelRepo: Optional[str]
    gpuType: Optional[str]
    numGpu: Optional[int]
    status: Optional[str]
    createdBy: Optional[Creator]
    endpointUrl: Optional[str]
    createdAt: Optional[datetime]
    updatedAt: Optional[datetime]
    phase: Optional[EndpointPhase] = None
    advancedConfig: Optional[EndpointAdvancedConfig] = None



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli User GQL Clients."""

from __future__ import annotations

from typing import Any, Dict, List

from friendli.client.graphql.base import GqlClient

GetCurrentUserInfoOp = """
query GetClientSession {
    clientSession {
        user {
            id
            name
            email
        }
    }
}
"""

GetUserTeamsOp = """
query GetClientTeams {
    clientSession {
        user {
            teams {
                edges {
                    node {
                        id
                        name
                        state
                        dedicated {
                            plan
                        }
                    }
                }
            }
        }
    }
}
"""

GetUserDefaultTeamOp = """
query GetDefaultTeam {
    clientUser {
        teamDefault {
            team {
                id
                name
                state
            }
        }
    }
}
"""


class UserGqlClient(GqlClient):
    """User gql client."""

    def get_current_user_info(self) -> Dict[str, Any]:
        """Get current user info."""
        response = self.run(query=GetCurrentUserInfoOp)
        return response["clientSession"]["user"]

    def get_teams(self) -> List[Dict[str, Any]]:
        """List user teams."""
        response = self.run(query=GetUserTeamsOp)
        return response["clientSession"]["user"]["teams"]["edges"]

    def get_team_ids(self) -> List[str]:
        """List team IDs."""
        response = self.run(query=GetUserTeamsOp)
        return [
            edge["node"]["id"]
            for edge in response["clientSession"]["user"]["teams"]["edges"]
        ]

    def get_default_team(self) -> Dict[str, Any]:
        """Get a default team."""
        response = self.run(query=GetUserDefaultTeamOp)
        return response["clientUser"]["teamDefault"]["team"]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI File System Management Utilities."""

from __future__ import annotations

import os
import re
from pathlib import Path
from typing import Union

friendli_directory = Path.home() / ".friendli"


def get_friendli_directory() -> Path:
    """Get a hidden path to Friendli config directory."""
    friendli_directory.mkdir(exist_ok=True)
    return friendli_directory


def storage_path_to_local_path(storage_path: str, source_path: Path) -> Path:
    """Translate a cloud storage path to the local file system path."""
    return Path(strip_storage_path_prefix(str(source_path / storage_path)))


def get_file_size(file_path: Union[str, Path]) -> int:
    """Calculate a file size in bytes.

    Args:
        file_path (Union[str, Path]): Path to the target file.

    Returns:
        int: The size of a file.

    """
    if isinstance(file_path, str):
        return os.stat(file_path).st_size
    return file_path.stat().st_size


def strip_storage_path_prefix(path: str) -> str:
    """Strip a prefix about auxiliary checkpoint info from the path.

    Args:
        path (str): Actual checkpoint storage path. The path may starts with the prefix
            that contain the iteration number and distributed training configuration.

    Returns:
        str: Path without the prefix.

    """
    return re.sub(
        pattern=r"iter_\d{7}/mp\d{3}-\d{3}pp\d{3}-\d{3}/",
        repl="",
        string=path,
    )


def attach_storage_path_prefix(
    path: Union[str, Path],
    iteration: int,
    mp_rank: int,
    mp_degree: int,
    pp_rank: int,
    pp_degree: int,
) -> str:
    """Attach a filename prefix to mark up iteration and dist info.

    Args:
        path (Union[str, Path]): The relative path to file.
        iteration (int): Checkpoint iteration number.
        mp_rank (int): Model parallelism(a.k.a. tensor parallelism) rank.
        mp_degree (int): Model parallelism(a.k.a. tensor parallelism) degree.
        pp_rank (int): Pipelined model parallelism rank.
        pp_degree (int): Pipelined model parallelism degree.

    Returns:
        str: Storage path with the prefix

    """
    # pylint: disable=line-too-long
    return f"iter_{iteration:07d}/mp{mp_rank:03d}-{mp_degree:03d}pp{pp_rank:03d}-{pp_degree:03d}/{path}"



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI URL Utilities."""

from __future__ import annotations

from urllib.parse import urljoin, urlparse

from typing_extensions import deprecated

discuss_url = "https://discuss.friendli.ai/"


def get_host(url: str) -> str:
    """Get a host part of a URL."""
    parsed_url = urlparse(url)
    scheme = parsed_url.scheme
    netloc = parsed_url.netloc
    return f"{scheme}://{netloc}/"


class URLProvider:
    """Service URL provider."""

    suite_url = ""
    api_url = ""
    training_url = ""
    registry_url = ""
    serving_url = ""
    auth_url = ""
    meter_url = ""
    observatory_url = ""
    web_backend_url = ""

    @classmethod
    def get_suite_uri(cls, path: str) -> str:
        """Get Friendli Suite URI."""
        return urljoin(cls.suite_url, path)

    @classmethod
    def get_auth_uri(cls, path: str) -> str:
        """Get PFA URI."""
        return urljoin(cls.auth_url, path)

    @classmethod
    def get_web_backend_uri(cls, path: str) -> str:
        """Get PF Web Backend API URI."""
        return urljoin(cls.web_backend_url, path)

    @classmethod
    @deprecated("All functionalities to be migrated to get_web_backend_uri")
    def get_training_uri(cls, path: str) -> str:
        """Get PFT URI."""
        return urljoin(cls.training_url, path)

    @classmethod
    def get_api_uri(cls, path: str) -> str:
        """Get PFT URI."""
        return urljoin(cls.api_url, path)

    @classmethod
    def get_serving_uri(cls, path: str) -> str:
        """Get PFS URI."""
        return urljoin(cls.serving_url, path)

    @classmethod
    def get_mr_uri(cls, path: str) -> str:
        """Get PFR URI."""
        return urljoin(cls.registry_url, path)

    @classmethod
    def get_meter_uri(cls, path: str) -> str:
        """Get PFM URI."""
        return urljoin(cls.meter_url, path)

    @classmethod
    def get_observatory_uri(cls, path: str) -> str:
        """Get PFO URI."""
        return urljoin(cls.observatory_url, path)


class ProductionURLProvider(URLProvider):
    """Production service URL provider."""

    suite_url = "https://suite.friendli.ai/"
    registry_url = "https://modelregistry.friendli.ai/"
    serving_url = "https://serving.friendli.ai/"
    auth_url = "https://auth.friendli.ai/"
    meter_url = "https://metering.friendli.ai/"
    observatory_url = "https://observatory.friendli.ai/"
    web_backend_url = "https://suite.friendli.ai/"
    training_url = "https://training.friendli.ai/api/"


class StagingURLProvider(URLProvider):
    """Staging service URL provider."""

    suite_url = "https://suite-staging.friendli.ai/"
    registry_url = "https://pfmodelregistry-staging.friendli.ai/"
    serving_url = "https://pfs-staging.friendli.ai/"
    auth_url = "https://pfauth-staging.friendli.ai/"
    meter_url = "https://pfmeter-staging.friendli.ai/"
    observatory_url = "https://pfo-staging.friendli.ai/"
    web_backend_url = "https://api-staging.friendli.ai/"
    training_url = "https://api-staging.friendli.ai/api/"


class DevURLProvider(URLProvider):
    """Dev service URL provider."""

    suite_url = "https://suite-dev.friendli.ai/"
    registry_url = "https://pfmodelregistry-dev.friendli.ai/"
    serving_url = "https://pfs-dev.friendli.ai/"
    auth_url = "https://pfauth-dev.friendli.ai/"
    meter_url = "https://pfmeter-dev.friendli.ai/"
    observatory_url = "https://pfo-dev.friendli.ai/"
    web_backend_url = "https://api-dev.friendli.ai/"
    training_url = "https://api-dev.friendli.ai/api/"



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Version utils."""


from __future__ import annotations

import importlib
import json
from importlib.metadata import version
from typing import List
from urllib.request import urlopen

from packaging.version import _BaseVersion
from packaging.version import parse as parse_version

FRIENDLI_PACKAGE_NAME = "friendli-client"
PYPI_BASE_URL = "https://pypi.org/pypi"


def is_latest_version(ver: str) -> bool:
    """Check if the installed CLI version is the latest release."""
    version = parse_version(ver)
    latest_version = get_latest_version()

    return version >= latest_version


def get_latest_version() -> _BaseVersion:
    """Get the latest CLI release version."""
    with urlopen(f"{PYPI_BASE_URL}/{FRIENDLI_PACKAGE_NAME}/json") as resp:
        pypi_info = json.loads(resp.read())
    latest_ver_string = pypi_info["info"]["version"]
    return parse_version(latest_ver_string)


def get_installed_version() -> str:
    """Get the currently installed CLI version."""
    return version(FRIENDLI_PACKAGE_NAME)


def check_dependencies_installed(deps: List[str]) -> bool:
    """Check if dependencies are installed."""
    for dep in deps:
        try:
            importlib.import_module(dep)
        except ImportError:
            return False

    return True



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

"""Friendli Client."""

from __future__ import annotations

import os
from typing import Optional, Union

import grpc
import grpc.aio
import httpx

import friendli
from friendli.client.graphql.endpoint import EndpointGqlClient
from friendli.client.graphql.model import ModelGqlClient
from friendli.sdk.api.chat.chat import AsyncChat, Chat
from friendli.sdk.api.completions import AsyncCompletions, Completions
from friendli.sdk.api.images.images import AsyncImages, Images
from friendli.sdk.resource.endpoint import EndpointApi
from friendli.sdk.resource.model import ModelApi

INFERENCE_ENDPOINT_URL = "https://inference.friendli.ai"


class FriendliClientBase:
    """Base class of Friendli client."""

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        team_id: Optional[str] = None,
        project_id: Optional[str] = None,
        use_dedicated_endpoint: bool = False,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
        use_grpc: bool = False,
        http_client: Optional[Union[httpx.Client, httpx.AsyncClient]] = None,
        grpc_channel: Optional[Union[grpc.Channel, grpc.aio.Channel]] = None,
    ):
        """Initializes FriendliClientBase."""
        if token is not None:
            friendli.token = token
        if team_id is not None:
            friendli.team_id = team_id
        if project_id is not None:
            friendli.project_id = project_id
        self._use_dedicated_endpoint = use_dedicated_endpoint
        self._base_url = base_url
        self._use_protobuf = use_protobuf

        if use_grpc:
            if base_url is None and grpc_channel is None:
                raise ValueError(
                    "One of `base_url` and `grpc_channel` should be set when `use_grpc=True`."
                )
            if http_client is not None:
                raise ValueError("You cannot use HTTP client when `use_grpc=True`.")
            if use_dedicated_endpoint:
                raise ValueError(
                    "`use_grpc=True` is not allowed for dedicated endpoints."
                )
        else:
            if grpc_channel is not None:
                raise ValueError(
                    "Setting `use_grpc=True` is required when `grpc_channel` is set."
                )
            if base_url is None:
                self._base_url = INFERENCE_ENDPOINT_URL

                if use_dedicated_endpoint:
                    self._base_url = os.path.join(self._base_url, "dedicated")


class Friendli(FriendliClientBase):
    """Friendli API client."""

    completions: Completions
    chat: Chat
    images: Images
    endpoint: EndpointApi
    model: ModelApi

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        team_id: Optional[str] = None,
        project_id: Optional[str] = None,
        use_dedicated_endpoint: bool = False,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
        use_grpc: bool = False,
        http_client: Optional[httpx.Client] = None,
        grpc_channel: Optional[grpc.Channel] = None,
    ):
        """Initializes Friendli."""
        super().__init__(
            token=token,
            team_id=team_id,
            project_id=project_id,
            use_dedicated_endpoint=use_dedicated_endpoint,
            base_url=base_url,
            use_protobuf=use_protobuf,
            use_grpc=use_grpc,
            http_client=http_client,
            grpc_channel=grpc_channel,
        )

        self.completions = Completions(
            base_url=self._base_url,
            use_protobuf=use_protobuf,
            use_grpc=use_grpc,
            http_client=http_client,
            grpc_channel=grpc_channel,
        )
        self.chat = Chat(
            base_url=self._base_url,
            use_protobuf=use_protobuf,
            use_grpc=use_grpc,
            http_client=http_client,
            grpc_channel=grpc_channel,
        )
        self.images = Images(
            base_url=self._base_url,
            http_client=http_client,
        )

        endpoint_client = EndpointGqlClient()
        model_client = ModelGqlClient()
        self.endpoint = EndpointApi(client=endpoint_client)
        self.model = ModelApi(client=model_client)

    def __enter__(self) -> Friendli:
        """Enter the context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the context manager and close resources."""
        self.close()

    def close(self) -> None:
        """Clean up all clients' resources."""
        self.completions.close()
        self.chat.close()
        self.images.close()


class AsyncFriendli(FriendliClientBase):
    """Async Friendli API client."""

    completions: AsyncCompletions
    chat: AsyncChat
    images: AsyncImages

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        team_id: Optional[str] = None,
        project_id: Optional[str] = None,
        use_dedicated_endpoint: bool = False,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
        use_grpc: bool = False,
        http_client: Optional[httpx.AsyncClient] = None,
        grpc_channel: Optional[grpc.aio.Channel] = None,
    ):
        """Initializes AsyncFriendli."""
        super().__init__(
            token=token,
            team_id=team_id,
            project_id=project_id,
            use_dedicated_endpoint=use_dedicated_endpoint,
            base_url=base_url,
            use_protobuf=use_protobuf,
            use_grpc=use_grpc,
            http_client=http_client,
            grpc_channel=grpc_channel,
        )

        self.completions = AsyncCompletions(
            base_url=self._base_url,
            use_protobuf=use_protobuf,
            use_grpc=use_grpc,
            http_client=http_client,
            grpc_channel=grpc_channel,
        )
        self.chat = AsyncChat(
            base_url=self._base_url,
            use_protobuf=use_protobuf,
            use_grpc=use_grpc,
            http_client=http_client,
            grpc_channel=grpc_channel,
        )
        self.images = AsyncImages(
            base_url=self._base_url,
            http_client=http_client,
        )

    async def __aenter__(self) -> AsyncFriendli:
        """Enter the asynchronous context manager."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the asynchronous context manager and close resources."""
        await self.close()

    async def close(self) -> None:
        """Clean up all clients' resources."""
        await self.completions.close()
        await self.chat.close()
        await self.images.close()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Client Dependency Injector."""

from __future__ import annotations

from contextlib import contextmanager
from typing import Callable, Iterable, Iterator, Optional, Type, Union

from injector import Binder, Injector, Module

_InstallableModuleType = Union[Callable[[Binder], None], Module, Type[Module]]

default_injector = Injector()
test_injector = None


def get_injector() -> Injector:
    """Gets injector for DI."""
    return test_injector or default_injector


@contextmanager
def patch_modules(
    modules: Union[
        _InstallableModuleType, Iterable[_InstallableModuleType], None
    ] = None,
    override_defaults: bool = False,
) -> Iterator[None]:
    """Patches injector modules."""
    global test_injector  # pylint: disable=global-statement
    old_injector = test_injector

    parent_injector: Optional[Injector] = None
    if old_injector:
        parent_injector = old_injector
    elif override_defaults:
        parent_injector = default_injector
    else:
        parent_injector = None

    test_injector = Injector(modules, parent=parent_injector)
    try:
        yield
    finally:
        test_injector = old_injector


def set_default_modules(
    modules: Union[
        _InstallableModuleType, Iterable[_InstallableModuleType], None
    ] = None,
):
    """Set default DI modules."""
    global default_injector  # pylint: disable=global-statement
    default_injector = Injector(modules)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Project CLI."""

from __future__ import annotations

import typer

from friendli.client.graphql.team import TeamGqlClient
from friendli.context import (
    get_current_project_id,
    get_current_team_id,
    set_current_project_id,
)
from friendli.formatter import PanelFormatter, TableFormatter
from friendli.utils.decorator import check_api
from friendli.utils.format import secho_error_and_exit

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=True,
)
project_table_formatter = TableFormatter(
    name="Project", fields=["node.id", "node.name"], headers=["ID", "Name"]
)
project_panel_formatter = PanelFormatter(
    name="Project Detail",
    fields=["pf_group_id", "id", "name"],
    headers=["Organization ID", "Project ID", "Name"],
)
member_table_formatter = TableFormatter(
    name="Members",
    fields=["id", "name", "email", "access_level"],
    headers=["ID", "Name", "Email", "Role"],
)


# pylint: disable=redefined-builtin
@app.command()
@check_api
def list():
    """List all accessible projects."""
    client = TeamGqlClient()

    team_id = get_current_team_id()
    if team_id is None:
        secho_error_and_exit(
            "Team is not configured. Set 'FRIENDLI_TEAM' environment variable, or use "
            "a CLI command 'friendli team switch'."
        )
    projects = client.get_projects(team_id=team_id)
    current_project_id = get_current_project_id()

    for project in projects:
        if (
            current_project_id is not None
            and project["node"]["id"] == current_project_id
        ):
            project["node"]["id"] = f"[bold green]* {project['node']['id']}"
            project["node"]["name"] = f"[bold green]{project['node']['name']}"
        else:
            project["node"]["id"] = f"  {project['node']['id']}"

    project_table_formatter.render(projects)


@app.command()
@check_api
def switch(
    project_id: str = typer.Argument(
        ...,
        help="ID of project to switch.",
    )
):
    """Switch currnet project context to run as."""
    client = TeamGqlClient()
    team_id = get_current_team_id()
    if team_id is None:
        secho_error_and_exit(
            "Team context should be set prior to setting project context"
        )

    accessible_project_ids = client.get_project_ids(team_id=team_id)
    if project_id not in accessible_project_ids:
        secho_error_and_exit(f"'{project_id}' is not valid team ID.")

    set_current_project_id(project_id)
    typer.secho(
        f"Project context is switched to '{project_id}'.",
        fg=typer.colors.GREEN,
    )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Decorator utils."""

from __future__ import annotations

import functools
from typing import Any, Callable

from friendli.errors import APIError, AuthenticationError, AuthorizationError
from friendli.utils.format import secho_error_and_exit


def check_api(func: Callable[..., Any]) -> Callable[..., Any]:
    """Check common API exceptions."""

    @functools.wraps(func)
    def inner(*args, **kwargs) -> Any:
        try:
            return func(*args, **kwargs)
        except (APIError, AuthorizationError, AuthenticationError) as exc:
            secho_error_and_exit(str(exc))

    return inner



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# pylint: disable=redefined-builtin, too-many-locals, too-many-arguments, line-too-long

"""Friendli Model CLI."""

from __future__ import annotations

import typer

from friendli.formatter import TableFormatter
from friendli.sdk.client import Friendli
from friendli.utils.compat import model_dump
from friendli.utils.decorator import check_api

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=True,
)

table_formatter = TableFormatter(
    name="Models",
    fields=[
        "id",
        "name",
    ],
    headers=[
        "ID",
        "Name",
    ],
)


@app.command("list")
@check_api
def list_models():
    """List models."""
    client = Friendli()
    models = client.model.list()
    models_ = [model_dump(model) for model in iter(models)]
    table_formatter.render(models_)



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

from __future__ import annotations

from typing import ClassVar as _ClassVar
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper

DESCRIPTOR: _descriptor.FileDescriptor

class ResponseFormat(_message.Message):
    __slots__ = ("type", "schema")

    class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        text: _ClassVar[ResponseFormat.Type]
        json_object: _ClassVar[ResponseFormat.Type]
        regex: _ClassVar[ResponseFormat.Type]
    text: ResponseFormat.Type
    json_object: ResponseFormat.Type
    regex: ResponseFormat.Type
    TYPE_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    type: ResponseFormat.Type
    schema: str
    def __init__(
        self,
        type: _Optional[_Union[ResponseFormat.Type, str]] = ...,
        schema: _Optional[str] = ...,
    ) -> None: ...



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Modules that configures bindings to providers."""

from __future__ import annotations

from injector import Binder, Module

from friendli import settings
from friendli.utils import url


class SettingsModule(Module):
    """Settings module."""

    def configure(self, binder: Binder) -> None:
        """Configures bindings for settings."""
        binder.bind(url.URLProvider, to=url.ProductionURLProvider)  # type: ignore
        binder.bind(settings.Settings, to=settings.ProductionSettings)  # type: ignore


default_modules = [SettingsModule]



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

"""Friendli Chat API."""

from __future__ import annotations

from typing import Optional

import grpc
import grpc.aio
import httpx

from friendli.sdk.api.chat.completions import AsyncCompletions, Completions


class Chat:
    """Chat API."""

    completions: Completions

    def __init__(
        self,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
        use_grpc: bool = False,
        http_client: Optional[httpx.Client] = None,
        grpc_channel: Optional[grpc.Channel] = None,
    ) -> None:
        """Initializes Chat."""
        self.completions = Completions(
            base_url=base_url,
            use_protobuf=use_protobuf,
            use_grpc=use_grpc,
            http_client=http_client,
            grpc_channel=grpc_channel,
        )

    def close(self) -> None:
        """Clean up all clients' resources."""
        self.completions.close()


class AsyncChat:
    """Asynchronous chat API."""

    completions: AsyncCompletions

    def __init__(
        self,
        base_url: Optional[str] = None,
        use_protobuf: bool = False,
        use_grpc: bool = False,
        http_client: Optional[httpx.AsyncClient] = None,
        grpc_channel: Optional[grpc.aio.Channel] = None,
    ) -> None:
        """Initializes AsyncChat."""
        self.completions = AsyncCompletions(
            base_url=base_url,
            use_protobuf=use_protobuf,
            use_grpc=use_grpc,
            http_client=http_client,
            grpc_channel=grpc_channel,
        )

    async def close(self) -> None:
        """Clean up all clients' resources."""
        await self.completions.close()
